/**********************************************
 ICE Proprietary Software - do NOT disseminate
 **********************************************/
/*
  ICE HyperLink Input/Output Controllers

  Jeff Schoen
  Innovative Computer Engineering, Inc.
  1/15/2003

  IO=SS

  Input: 210Mby each channel
   iclkp = input clock
   iclkn = A/B valid
   idatp = A/B  7:0
   idatn = A/B	15:8

  Output:	100Mby each channel
   oclkp = A valid
   oclkn = B valid
   odatp = A  7:0
   odatn = B  7:0

  IO=HH

  Input: 800Mby each channel
   iclkp = input clock
   iclkn = input clock
   idatp = A/B  7:0
   idatn = A/B	15:8

  Output: 800Mby each channel
   oclkp = output clock
   oclkn = output clock
   odatp = A/B  7:0
   odatn = B/B  15:8

  IO=RR

  Input: 4Gby each 4-lane channel
  Output: 4Gby each 4-lane channel
  ICEt transport guarantees 8*N byte align

  Packet Header: 0xccssmmA5
    A5 = start symbol
    mm = command flags [3:0]=ohold
    ss = size in 64by chunks
    cc = PM channel number or IOM timecode

  Test Pattern: 

    FF:00 EE:11 DD:22 CC:33   BB:44 AA:55 99:66 88:77   77:88 66:99 55:AA 44:BB   33:CC 22:DD 11:EE 00:FF

    Bit#0 0110011001100110 0110011001100110
    Bit#2 0101101001011010 0101101001011010
    Bit#3 0101010110101010 0101010110101010
    Bit#4 0101010101010101 1010101010101010

*/
`include "mdefs.h"
`include "dmadefs.h"
`include "ramdefs.h"
`include "hypdefs.h"

// we swapped P/N on bit 2 for layout reasons
`ifdef V6M
`define IFLP 8'h00 
`define OFLP 8'h04
`else
`define IFLP 8'h00 
`define OFLP 8'h00 
`endif

`ifdef S6M
`define HYP_PIPE 1
`define HYP_IDLY 1
`else
`define HYP_PIPE 0
`define HYP_IDLY 0
`endif

`ifndef HYP_PKT
`define HYP_PKT `DMA_PKT_LEN
`endif

module hyperlink ();
endmodule

module hypin (
`ifdef HYP_RIO
  iclk,irst,iena,ibus,
`else
  iclk,iclks, idatp,idatn,
`endif
  ihold,ohold,
  ioclk,rio, ostatha,oenaha,oselha,obusha,  ostathb,oenahb,oselhb,obushb, 
  mode,route,calib, rha,rhb,test);

  parameter IBW=32;
  parameter MODE=5;
  parameter TUNERS=1;
  parameter FLGA=0;
  parameter FLGB=(MODE==4)?`CINV:0;

  localparam NR=(MODE>7)?16:2;
  localparam ND=(MODE>7)?3:5;
  localparam IDLY=`HYP_IDLY;
  localparam FLGO=(`HYP_PIPE!=0)? `PACKET|`OREG : `PACKET;

  localparam DBW=(MODE>7)? 64:32;
  localparam IBV=IBW/8;
  localparam DBV=DBW/8;
  localparam NPP=`HYP_PKT;

`ifdef HYP_RIO
  input iclk, irst, iena;
  input [DBW-1:0] ibus;
`else
  input iclk;
  input [3:0] iclks;
  input [7:0] idatp,idatn;
`endif
  output [3:0] ihold,ohold;

  input ioclk,rio,oenaha,oenahb;
  output ostatha,ostathb;
  output [7:0] oselha,oselhb;
  output [IBW-1:0] obusha,obushb;

  input mode;
  input [31:0] route;
  output [31:0] calib;
  output rha,rhb;
  output [7:0] test;

  wire L=0, H=1;

  wire hirdya,hirdyb,hival,ha2td,hb2td,hcsm,hcab;
  wire [DBW-1:0] hidat; 
  wire [DBV-1:0] hisel; 
  reg ha2t,hb2t,ha2x,hb2x,rha,rhb,iholda,iholdb,hirdyad,hirdybd;
  delaypipe #(15,1) dpa (ioclk,ha2t,ha2td);
  delaypipe #(15,1) dpb (ioclk,hb2t,hb2td);
  always @(posedge ioclk) begin
    ha2t <= (route[`HA2TA] || route[`HA2TB]) && (TUNERS==1);
    hb2t <= (route[`HB2TA] || route[`HB2TB]) && (TUNERS==1);
    ha2x <= (route[`HA2TA] || route[`HA2TB] || route[`HA2CA] || route[`HA2CB] || route[`HA2R]); 
    hb2x <= (route[`HB2TA] || route[`HB2TB] || route[`HB2CA] || route[`HB2CB] || route[`HB2R]); 
    rha  <= rio || (ha2t^ha2td);
    rhb  <= rio || (hb2t^hb2td);
  end
  always @(posedge iclk) begin
    hirdyad <= hirdya;
    hirdybd <= hirdyb;
    iholda  <= ha2x && !hirdya && !hirdyad;
    iholdb  <= hb2x && !hirdyb && !hirdybd;
  end
  wire [7:0] vtest;
`ifdef HYP_RIO
  hypinr #(DBW,DBW,NPP) qhi (iclk,H, iena,ibus, hival,hisel,hidat, mode,ohold,hcsm,hcab,calib,vtest);
`else
  wire [31:0] vdat;
  wire vclk=iclk, pclk=iclks[0], prst=iclks[1], ss=iclks[2];
  diqdrN #(8,1,IDLY,`IFLP) qdat (idatp,idatn, vclk,pclk,prst,ss, vdat[7:0],vdat[15:8],vdat[23:16],vdat[31:24]);
  hypinq #(DBW,DBW,1,`HYP_ONOF) qhi (vclk,H,vdat, hival,hisel,hidat, mode,ohold,hcsm,hcab,calib,vtest);
`endif
  // input FIFOs
  fifoNxM #(DBW+DBV,IBW+IBV,NR,ND,`BPAR|FLGA,FLGO) fa (iclk,rha,hirdya,hival&!hcab,{hisel,hidat},  ioclk,rha,ostatha,oenaha,{oselha,obusha});
  fifoNxM #(DBW+DBV,IBW+IBV,NR,ND,`BPAR|FLGB,FLGO) fb (iclk,rhb,hirdyb,hival& hcab,{hisel,hidat},  ioclk,rhb,ostathb,oenahb,{oselhb,obushb});
  // input hold logic
  assign ihold = {L,iholdb,iholda,iholda||iholdb};

//  assign test = {iclk,oenahb,oenaha,ostatha, hcab,hisel[0],hival,iena};
  assign test = {ihold,hival,hirdya};

endmodule

module hypout (
`ifdef HYP_RIO
  oclk,orst,ordy, oena,obus,
`else
  oclk,oclks, odatp,odatn,
`endif
  ihold,ohold,
  ioclk,rio, istatha,ienaha,iselha,ibusha, istathb,ienahb,iselhb,ibushb, 
  istat1,istat2, route,speed,test);

  parameter IBW=32;		// IO bus width in bits
  parameter MODE=5;

  localparam DBW=(MODE>7)? 64:32;
  localparam IBV=IBW/8;
  localparam DBV=DBW/8;
  localparam NW=(IBW/DBW);	// IO bus width in words
  localparam NPP=`HYP_PKT;
  localparam NR=(MODE>7)?8:4;
  localparam ND=(MODE>7)?3:3;

  wire L=0, H=1;

`ifdef HYP_RIO
  input oclk,orst,ordy;
  output oena;
  output [DBW-1:0] obus;
`else
  input oclk;
  input [3:0] oclks;
  output [7:0] odatp,odatn;
`endif
  input [3:0] ihold,ohold;

  input ioclk,rio,ienaha,ienahb;
  output istatha,istathb;
  output istat1,istat2;
  input [7:0] iselha,iselhb;
  input [IBW-1:0] ibusha,ibushb;

  input [31:0] route;
  input [7:0] speed;		// speed control 200/(N+2) link port
  output [7:0] test;

  wire ha2x = (route[`HA2TA] || route[`HA2TB] || route[`HA2CA] || route[`HA2CB] || route[`HA2R]); 
  wire hb2x = (route[`HB2TA] || route[`HB2TB] || route[`HB2CA] || route[`HB2CB] || route[`HB2R]); 

  reg ivalhx; always @(posedge ioclk) ivalhx <= (ienaha || ienahb);
  reg [2:0] ivals; always @(posedge ioclk) ivals <= {ienahb && !iselha[7], ienaha && !iselha[7], (ienaha || ienahb) && iselha[7]};
  reg iselx;  always @(posedge ioclk) iselx  <= !iselx && ivalhx;

  wire [IBV-1:0] iselhx = (IBW>32)? {iselha,iselha} : iselx? iselha[IBV+IBV-1:IBV] : iselha[IBV-1:0];
  wire [DBV-1:0] hosel;
  wire [2:0] hordys,istats;
  wire [7:0] vtest;
  wire hoena,hordy;
  wire [DBW-1:0] hodat,vdat; // note that ibusha/iselha are the same as ibushb/iselhb

`ifdef S6M
  assign hordy = (hordys[0] && !ohold[0]);	// no multichannel fifo since there are no COREs, only MCORES on S6M
  fifoNxM #(IBW+IBV,DBW+DBV,2,3,`BPAR,`PACKET) fab (ioclk,rio,istats[0],ivalhx,{iselhx,ibusha}, oclk,rio,hordys[0],hoena,{hosel,hodat});
`else
  reg [1:0] hochn; always @(posedge oclk) hochn <= (hordys[2]&&!ohold[2])? 2 : (hordys[1]&&!ohold[1])? 1 : (hordys[0]&&!ohold[0])? 0 : 3;
  reg [1:0] hocho; always @(posedge oclk) if (!hoena && hochn!=3) hocho <= hochn;
  wire [2:0] hoenas = { hoena&&(hocho==2), hoena&&(hocho==1), hoena&&(hocho==0) };
  assign hordy = (hochn!=3);
  mcfifoNxM #(IBW+IBV,DBW+DBV,NR,ND,`BPAR,(NPP>`DMA_PKT_LEN)?0:`PACKET) fab (ioclk,rio,istats,ivals,{iselhx,ibusha}, oclk,rio,hordys,hoenas,{hosel,hodat});
`endif

`ifdef HYP_RIO
  wire aohold = (ohold!=0);
  hypoutr #(DBW,DBW,NPP) qho (oclk,ordy, oena,obus, hordy,hoena,hosel,hodat, {aohold,ihold},vtest);
  assign test = {hordy,ordy,oena,hoena,ohold[0]};
`else
  wire vclk=oclk, pclk=oclks[0], prst=oclks[1], ss=oclks[2];
  hypoutq #(DBW,DBW,NPP,`HYP_ONOF) qho (vclk,H,vdat, hordy,hoena,hosel,hodat, ihold,vtest);
  doqdrN #(8,1,`OFLP) qdat (odatp,odatn, vclk,pclk,prst,ss,L, vdat[7:0], vdat[15:8], vdat[23:16], vdat[31:24]);
  assign test = {hordy,hoena,ohold[0]};
`endif

  assign istatha=istats[0], istathb=istats[0], istat1=istats[1], istat2=istats[2]; 

endmodule



/*
                   idat jdat kdat
                   A5
  isyna            00   A5
  isync            01   00   A5
4 icycle           02   01   00
3 icycle           03   02   01
2 icycle           A5   03   02
1 icycle   idone   00   A5   03
  isync            01        A5

  iena - idat valid - qualify isyna,isynb
  jena - jdat valid - qualify isync,icycle
  kena - kdat valid - qualify output
*/
module hypinq (iclk,iena,idat, hval,hsel,hdat, modes,hold,csm,cab,calib,test);
  parameter IBW=32;
  parameter OBW=32;
  parameter NPKT=1;
  parameter FLG=0;

  localparam OBV=OBW/8;
  localparam IBV=IBW/8;

  input iclk,iena;
  input [IBW-1:0] idat;
  output hval;
  output [OBV-1:0] hsel;
  output [OBW-1:0] hdat;
  input  [2:0] modes;
  output [3:0] hold;
  output csm,cab;
  output [7:0] calib;
  output [7:0] test;

  wire L=0, H=1;
  wire mode = modes[0];
  wire fsync = modes[1];
  wire fixup = modes[2];

  wire [31:0] idatx = fixup? {idat[31:27],idat[30],idat[25:24], idat[23:19],idat[22],idat[17:16],
                              idat[15:11],idat[14],idat[9:8], idat[7:3],idat[6],idat[1:0]} : idat;

  // packet logic
  reg syt0,syt1,syt0d,syt1d,oop,active;
  reg isyna,isynb,isync,icycle,idone,ipha;
  reg [31:0] jdat,kdat;
  reg [7:0] icnt,ichan;
  reg [3:0] hold;
  reg sen,ten,chm,chb;
  wire [3:0] npkt = (NPKT>0)? NPKT : fixup? 4'h4 : kdat[19:16];
  wire [7:0] vchan = kdat[31:24];
  wire vchb; vote3 v1 (vchb, vchan[4],vchan[2],vchan[0]);
  wire vsen; vote3 v2 (vsen, vchan[5],vchan[3],vchan[1]);
  wire vten; vote3 v3 (vten, vchan[7],vchan[6],1'b0);
  wire vchm; vote3 v4 (vchm, vchan[7],vchan[6],1'b1);
  wire voteh; vote3 vh (voteh, vchan[7],vchan[6],vchan[5]);
  wire votel; vote3 vl (votel, vchan[2],vchan[1],vchan[0]);
  wire ichn = isync;
  always @(posedge iclk) begin
    syt0 <= (jdat[ 7: 0]==8'hAA); syt0d <= syt0 && !(isync || active);
    syt1 <= (jdat[23:16]==8'hAA); syt1d <= syt1 && !(isync || active);
    oop  <= syt1d? H : syt0d? L : oop;
    jdat <= idatx;
    kdat <= oop? {idatx[15:0],jdat[31:16]} : jdat;
    isyna  <= (idat[7:0]==8'hA5)   || (fixup && idat[7:0]==8'hA1);
    isynb  <= (idat[23:16]==8'hA5) || (fixup && idat[23:16]==8'hA1);
    isync  <= !isync && !(icycle && !idone) && (oop? isynb : isyna);
    icycle <= (isync && (fixup || kdat[18:16]!=0)) || (icycle && !idone) || mode;
    idone  <= icycle && (icnt[5:0]==2);
    if (isync) icnt <= {npkt,4'h0}; else if (icycle) icnt <= icnt-1; // 16x4by pkt count for NPKT 
    if (isync) ipha <= 0; else ipha <= !ipha;
    if (mode) chb <= 0; else if (ichn) chb <= vchb;
    if (mode) chm <= 0; else if (ichn) chm <= vchm;
    if (mode) sen <= 0; else if (ichn) sen <= vsen;
    if (mode) ten <= 0; else if (ichn) ten <= vten;
    if (mode) hold <= 0; else if (ichn) hold <= kdat[11:8];
    if (mode) ichan <= 0; else if (ichn) ichan <= voteh? {H,L,L,vchan[4:0]} : {L,{6{L}},votel};
  end

`ifdef ICE_MB
  wire [OBV-1:0] ksel = ipha? ichan[7:4] : ichan[3:0];	// need full input channel address
  assign cab  = ichan[0];
`else
  wire [OBV-1:0] ksel = {ten,sen,ten,sen};		// only A/B channel number and startup synchronization signals are important
  assign cab  = chb;
`endif

  reg [OBV-1:0] xdat;
  wire jcycle = isync || (icycle && !idone);
 generate
 if (FLG&`HYP_TC) begin:tc
  reg xsync,tcval,tcbit,tcbar,tcuse;
  reg [7:0] tcoff;
  always @(posedge iclk) begin
    if (isync) tcoff <= kdat[31:24];
    xsync <= (isync && kdat[11:8]==4'h1);
    tcval <= xsync || (tcval && (icnt[3:0]!=0));
    tcbit <= tcoff[ ~icnt[2:0] ];
    if (!tcval) tcbar <= 0; else tcbar <= ~icnt[2];	// 4 cyc start pulse
    if (!tcval) tcuse <= 0; else tcuse <= ~icnt[3];	// 8 cyc binary offset count
    xdat[0] <= tcuse?tcbit:tcbar;
    if (jcycle) xdat[1] <= fsync;
  end
 end
 else begin:ntc
  always @(posedge iclk) xdat <= 0;
 end
 endgenerate

  assign hval = icycle;
  assign hsel = ksel;
  assign hdat = {xdat,kdat};
  assign csm  = chm;

  // calibration check
  reg err,estr,eend,npkterr;
  reg [3:0] tst,errc;
  reg [7:0] tcnt,errs,terrs;
  wire [3:0] tst0 = {tst[3:1],L};
  wire [3:0] tst1 = {tst[3:1],H};
  always @(posedge iclk) begin
    if (isync|icycle) tcnt <= 0; else tcnt <= tcnt+1;
    estr <= (tcnt==5);
    eend <= (tcnt==255);
    if (estr) tst <= {jdat[19:17],L}; else tst <= tst+2;
    if (estr) errs <= 0; else if (err) errs <= errs+1;
    // err <= {~tst1,~tst1,tst1,tst1,~tst0,~tst0,tst0,tst0} != jdat;
    errc[0] <= kdat[ 7: 0] != { tst0, tst0};
    errc[1] <= kdat[15: 8] != {~tst0,~tst0};
    errc[2] <= kdat[23:16] != { tst1, tst1};
    errc[3] <= kdat[31:24] != {~tst1,~tst1};
    active  <= isync || icycle || (active && !tcnt[3]);
    if (active) err <= 0; else err <= (errc!=0);
    if (eend) terrs <= errs;
    if (ichn) npkterr <= (kdat[19:16] != `HYP_PKT);
  end
  assign calib = terrs;

  reg [7:0] aaa; always @(posedge iclk) if (!hold[0]) aaa <= 0; else aaa <= aaa+1;
  reg fff; always @(posedge iclk) fff <= !fff;
  reg ccc; always @(posedge iclk) ccc <= (terrs!=0);
//  assign test = {fff,err,estr,oop,mode,chb,icycle,isync};
//  assign test = {fff,err,estr,oop, ccc,mode,icycle,isync};
//  assign test = kdat[7:0];
//  assign test = {hold,oop,icycle,isync,fff};
  assign test = {ten,sen,chm,cab, err,icycle,npkterr,iena};

endmodule

/*
  ostart
  ostartd  ocyclep  hena
4 ocycle   ocyclep  hsel              hdat 
3 ocycle   ocyclep  hseld  oheader    tctick
2 ocycle   ocyclep  odone
1 ocycle            ostart

*/

module hypoutq (oclk,oena,odat, hrdy,hena,hsel,hdat, hold,test);
  parameter IBW=32;
  parameter OBW=32;
  parameter NPKT=1;
  parameter FLG=0;	// 1=OnOffProtocol 2=TimeCode

  localparam IBV=IBW/8;
  localparam OBV=OBW/8;

  input oclk;
  input oena;
  output [OBW-1:0] odat;

  input hrdy;
  output hena;
  input [IBV-1:0] hsel;
  input [IBW-1:0] hdat;

  input [3:0] hold;
  output [7:0] test;

  wire L=0, H=1;

  // output
  reg ostart,ocyclep,ocycle,ocycled,odone,ordyd,oenad,ahold,tctick,tcval;
  reg [7:0] ocnt,tcoff;
  reg [3:0] tst,hseld,ohold,lhold,jhold,tdatl;
  reg [OBW-1:0] obusa,obusb,obusc;
  wire [3:0] tst0 = {tst[3:1],L};
  wire [3:0] tst1 = {tst[3:1],H};
  wire [3:0] npkt = NPKT;
  wire [3:0] tdat = hdat[IBW-1:IBW-4];
  wire oheader; delaypipe #(3,1) dp (oclk,ostart,oheader);
  wire [7:0] hf1 = (FLG&`HYP_TC)? {7'd0,tcval} : {4'd0,jhold};
  wire [7:0] hf3 = (FLG&`HYP_TC)? tcoff : {hsel,hseld};
  wire uahold    = (FLG&`HYP_ONOF)? ahold : L;
  wire ordy      = hrdy;
  always @(posedge oclk) begin
    ordyd   <= ordy;
    ostart  <= oena && !ostart && !(ocyclep && !odone) && (ordy || uahold);
    ocyclep <= (ostart && ordyd) || (ocyclep && !odone);
    ocycle  <= ocyclep;
    ocycled <= ocycle;
    if (!ocycle) ocnt <= {npkt,4'h0}; else ocnt <= ocnt-1; // 16x4by pkt count for NPKT 
    odone   <= (ocnt[5:0]==3);	// bits good up through 256 by packet
    tst     <= tst+2;
    if (ocycle) obusa <= hdat;
    obusb   <= ocycled? obusa : {~tst1,~tst1, tst1,tst1, ~tst0,~tst0, tst0,tst0};
    if (ocycle) tdatl <= tdat;
    tctick  <= ocycle && (tdat!=0) && (tdatl==0); // OPPS rising edge
    tcval   <= tctick || (tcval && !oheader);
    if (tctick) tcoff <= {~ocnt[5:0], tdatl[0]? 2'd0 : tdatl[1]? 2'd1 : tdat[2]? 2'd2 : 2'd3 };
    obusc   <= oheader? {hf3, 4'h0,ocycled?npkt:4'h0, hf1, 8'hA5} : obusb;
    hseld   <= hsel;
    ohold   <= hold;
    jhold   <= ohold;
    lhold   <= oheader? jhold : lhold;
    ahold   <= oena && (jhold!=0 || lhold!=0 || !oenad); // if current or last sent is to hold off or initializing output
    oenad   <= oena;
  end
  assign hena = ocyclep;
  assign odat = obusc;

  reg fff; always @(posedge oclk) fff <= !fff;
  assign test = {hold, fff,ordy,ocycle,ostart};
//  assign test = {hseld,hold[0],ordy,ocycle,ostart};

endmodule


/*
 Hypertransport over RocketIO port code
 Error checking and synchronization happens in the RIO layer
 note: iena is not continuous

  iena idat
  jena jdat isyna
  kena kdat isync
            cycle

*/
module hypinr (iclk,ena, iena,idat, hval,hsel,hdat, modes,hold,csm,cab,calib,test);
  parameter IBW=64;
  parameter OBW=64;
  parameter NPKT=1;
  parameter FLG=0;

  localparam WIDE=(IBW==64)? 1:0;
  localparam OBV=OBW/8;
  localparam IBV=IBW/8;

  input iclk,ena,iena;
  input [IBW-1:0] idat;
  output hval;
  output [OBV-1:0] hsel;
  output [OBW-1:0] hdat;
  input  [1:0] modes;
  output [3:0] hold;
  output csm,cab;
  output [7:0] calib;
  output [7:0] test;

  wire L=0, H=1;
  wire mode = modes[0];
  wire fsync = modes[1];

  // packet logic
  reg isyna,isync,icycle,idone,jena,kena;
  reg sen,ten,chm,chb;
  reg [3:0] hold;
  reg [7:0] icnt,ichan;
  reg [IBW-1:0] jdat,kdat;
  wire [3:0] npkt = (NPKT>0)? NPKT : kdat[19:16];
  wire [7:0] ncnt = npkt<<(4-WIDE);
  wire [7:0] vchan = kdat[31:24];
  wire vchb = vchan[0];
  wire vsen = vchan[1];
  wire vten = vchan[6];
  wire vchm = vchan[7];
  wire ichn = jena && isync;
  always @(posedge iclk) begin
    jena <= iena; 
    kena <= jena;
    if (iena) jdat <= idat;
    if (jena) kdat <= jdat;
    if (iena) isyna  <= (idat[7:0]==8'hA5);
    if (jena) isync  <= isyna && !(icycle && !idone) && !isync;
    if (jena) icycle <= (ena && isync && kdat[18:16]!=0) || (icycle && !idone) || mode;
    if (jena) idone  <= icycle && (icnt[5:0]==2);
    if (ichn) icnt <= ncnt; else if (icycle&jena) icnt <= icnt-1; // 16x4by pkt count for NPKT 
    if (mode) chb <= 0; else if (ichn) chb <= vchb;
    if (mode) chm <= 0; else if (ichn) chm <= vchm;
    if (mode) sen <= 0; else if (ichn) sen <= vsen;
    if (mode) ten <= 0; else if (ichn) ten <= vten;
    if (mode) hold <= 0; else if (ichn) hold <= kdat[11:8];
    if (mode) ichan <= 0; else if (ichn) ichan <= vchm? {H,L,vchan[5:0]} : {L,{6{L}},vchb};
  end

`ifdef ICE_MB
  wire [OBV-1:0] ksel = ichan;			// need full input channel address
  assign cab  = ichan[0];
`else
  wire [OBV-1:0] ksel = {2{ten,sen,ten,sen}};	// only A/B channel number and startup synchronization signals are important
  assign cab  = chb;
`endif

  assign hval = kena&icycle;
  assign hsel = ksel;
  assign hdat = kdat;
  assign csm  = chm;

  assign calib = 0;
  wire err = 0;

  reg fff; always @(posedge iclk) if (!iena) fff <= !fff;
  assign test = {ten,sen,chm,cab, err,icycle,fff,iena};

endmodule

/*
  ostart
           ocyclep  hena
4 ocycle   ocyclep  hsel   oheader    hdat 
3 ocycle   ocyclep  hseld  ocycled    obusa	oena-h
2 ocycle   ocyclep  odone  	      obusc	oena-d
1 ocycle            ostart

4 ocycle

*/

module hypoutr (oclk,ena, oena,odat, hrdy,hena,hsel,hdat, hold,test);
  parameter IBW=64;
  parameter OBW=64;
  parameter NPKT=1;
  parameter FLG=0;	// 1=OnOffProtocol 2=TimeCode

  localparam WIDE=(IBW==64)? 1:0;
  localparam IBV=IBW/8;
  localparam OBV=OBW/8;
  localparam ONOF=(FLG&`HYP_ONOF)?1:0;

  input oclk,ena;
  output reg oena;
  output [OBW-1:0] odat;

  input hrdy;
  output hena;
  input [IBV-1:0] hsel;
  input [IBW-1:0] hdat;

  input [4:0] hold;
  output [7:0] test;

  wire L=0, H=1;

  // output
  reg ostart,ocyclep,ocycle,odone,ordy,ordyd,enad,ahold,ostartd,ocycled;
  reg [7:0] ocnt;
  reg [7:0] opcnt,oqcnt;
  reg [3:0] ohold,lhold,jhold;
  reg [OBW-1:0] obusa,obusc;
  wire [3:0] npkt = NPKT;
  wire [7:0] ncnt = NPKT<<(4-WIDE);
  wire [31:0] header = {hsel, 4'h0,ocycle?npkt:4'h0, 4'd0,jhold, 8'hA5};
  wire oheader; delaypipe #(2,1) dph (oclk,ostart,oheader);
  wire opad = oqcnt[7];
  wire aohold = hold[4];
  always @(posedge oclk) begin
    ordy    <= hrdy;
    ordyd   <= ordy;
//    ostart  <= ena && !ostart && !(ocyclep && !odone) && (ordy || ahold);
    ostart  <= ena && !ostart && !ocyclep && (ordy || ahold);	// write one cycle less so reciever can keep up
    ostartd <= ostart;
    ocyclep <= (ostart && ordyd) || (ocyclep && !odone);
    ocycle  <= ocyclep;
    ocycled <= ocycle;
    if (!ocycle) ocnt <= ncnt; else ocnt <= ocnt-1;	// 16x4by pkt count for NPKT 
    odone   <= (ocnt[5:0]==3);				// bits good up through 256 by packet
    if (!ocycle) obusa <= 0; else obusa <= hdat;
    obusc   <= oheader? {header,header} : obusa;
    ohold   <= hold;
    jhold   <= ohold;
    lhold   <= oheader? jhold : lhold;
    ahold   <= ONOF && ena && (jhold!=0 || lhold!=0 || !enad);// if current or last sent is to hold off or initializing output
    enad    <= ena;
//    oena    <= oheader || ocycled || (ONOF? opcnt[3] : ena&&opad );
//    oena    <= oheader || ocycled || (ena&&opcnt[3]) || (ena&&opad);
    oena    <= oheader || ocycled || (opcnt[3] && !aohold);
    if (oena)   opcnt <= 0; else opcnt <= opcnt+1;		// opcnt will leak zeros at 1/8 full rate if nothing to send
    if (ocycle) oqcnt <= ~0; else if (opad) oqcnt <= oqcnt-1;	// oqcnt will pad each output with up to 128 clocks of zeros
  end
  assign hena = ocyclep;
  assign odat = obusc;

  assign test = {hold, oena,ordy,ocycle,ena};

endmodule
