/**********************************************
 ICE Proprietary Software - do NOT disseminate
 **********************************************/
/*
  Various RAM / FIFO s

  Fifo write valid on iena
  Fifo read valid one cycle after oena

  Parameter DP is the number of 2Kby DualPort Block RAMS to use.
  If DP=0, it will use 1 or 2 as necessary to cover the larger of the two data bus widths (32b per RAM Interface).

  Parameter 2**BM is number of buffer splits, the ready flags guarantee 2 buffer splits available for read or write
  Thus a 4Kby fifo with BM=3 will always have 4K/8 * 2 or 1K of data available when flags are good

  Flags:
  CINV  Clock Invert
  CENA  Clock Enable
  OREG  Registered Output Buffer
  TIGHT Tight buffer coupling
  PACKET One 64-byte packet buffering

  Jeff Schoen
  Innovative Computer Engineering, Inc.
  1/15/2005

*/

`include "mdefs.h"
`include "../lib/ramdefs.h"

// Block RAM architecture specifics
`ifdef verilator
`define USERAMBXX
//`define DEBUG
// uncomment this to emulate Altera
//`define NO16BRAM
// uncomment this to emulate Virtex4-
//`define NO32BRAM

`elsif VIRTEX2
`define NO32BRAM
`define NORAMREG
`endif

`ifdef VIRTEX4
`define NO32BRAM
`endif

`ifdef VIRTEX5
`define NO32BRAM
`endif

`ifdef SPARTAN6
`define NO32BRAM
`endif

`ifdef ALTERA
`define RAMBW 16
`define NO32BRAM
`else
`define RAMBW 32
`endif

`ifdef STRATIX4
`define NO16BRAM
`endif

// Block RAM default kbytes and addr bits
`ifdef NO16BRAM
`define RAMKB 1
`elsif NO32BRAM
`define RAMKB 2
`else
`define RAMKB 4
`endif

`ifndef SYS_CLKF
`define SYS_CLKF 100
`endif

`ifdef DMA_PKT_4X
`define DMA_PKT_LEN 4
`define LPKTS 11
`elsif DMA_PKT_2X
`define DMA_PKT_LEN 2
`define LPKTS 10
`else
`define DMA_PKT_LEN 1
`define LPKTS 9
`endif

module rams();

endmodule

module fiaoNxM (iclk,irst, irdy,ival,idat, 
                oclk,orst, ordy,oena,odma,onxt,oadr,olen,odat);
  parameter ND=8;
  parameter MD=8;
  parameter NR=0;
  parameter BM=3;
  parameter NFLG=0;
  parameter MFLG=0;

  localparam ILEN=1024;

  input irst, iclk, ival;
  input orst, oclk, oena, onxt;
  input [31:0] odma;
  input [15:0] oadr,olen;
  output irdy, ordy;

  input [ND-1:0] idat;
  output [MD-1:0] odat;

  // address logic for stream crossbar input 
  wire inxt = 0; wire [5:0] iadr = (ND/8); wire [15:0] ilen = ILEN; wire [31:0] idma = 0;

  aiaoNxM #(ND,MD,NR,BM,NFLG|`BSTREAM,MFLG) bif (
      iclk,irst, irdy,ival,idma,inxt,iadr,ilen,idat,
      oclk,orst, ordy,oena,odma,onxt,oadr,olen,odat);

endmodule


module aifoNxM (iclk,irst, irdy,ival,idma,inxt,iadr,ilen,idat, 
                oclk,orst, ordy,oena,odat);
  parameter ND=8;
  parameter MD=8;
  parameter NR=0;
  parameter BM=3;
  parameter NFLG=0;
  parameter MFLG=0;

  localparam OLEN=256;

  input irst, iclk, ival, inxt;
  input orst, oclk, oena;
  input [31:0] idma;
  input [15:0] iadr,ilen;
  output irdy, ordy;

  input [ND-1:0] idat;
  output [MD-1:0] odat;

  // address logic for stream crossbar output
  wire onxt = 0; wire [5:0] oadr = MD/8; wire [15:0] olen = OLEN; wire [31:0] odma = 0;

  aiaoNxM #(ND,MD,NR,BM,NFLG,MFLG|`BSTREAM) bif (
      iclk,irst, irdy,ival,idma,inxt,iadr,ilen,idat,
      oclk,orst, ordy,oena,odma,onxt,oadr,olen,odat);

endmodule


module aiaoNxM (vclk,vrst, vrdy,vval,vdma,vnxt,vadr,vlen,vdat, 
                wclk,wrst, wrdy,wena,wdma,wnxt,wadr,wlen,wdat);
  parameter ND=8;	// data width in bits
  parameter MD=8;	// data width in bits
  parameter NR=0;	// number of 1K rams
  parameter BX=4;	// address bits in comparison 
  parameter NFLG=0;	// N flags
  parameter MFLG=0;	// M flags

  localparam NKR=(NR>0)? NR : (ND>288 || MD>288)? 32 : (ND>144 || MD>144)? 16 : (ND>72 || MD>72)? 8 : (ND>36 || MD>36)? 4 : 2;
  localparam LRS=(NKR>128)? 21 : (NKR>64)? 20 : (NKR>32)? 19 : (NKR>16)? 18 : (NKR>8)? 17 : (NKR>4)? 16 : (NKR>2)? 15 : 14; // log2 of ram size
  localparam LND=(ND>=128)? 4 : (ND>=64)? 3 : (ND>=32)? 2 : (ND>=16)? 1 : 0; 	// LND|LMD is log2 of port widths in bytes
  localparam LMD=(MD>=128)? 4 : (MD>=64)? 3 : (MD>=32)? 2 : (MD>=16)? 1 : 0;
  localparam NB=(1<<LND);
  localparam MB=(1<<LMD);
  localparam NA=LRS-3;
  localparam MA=LRS-3;
  localparam LA=NA;
  localparam BM=(BX>0)?BX:LA;

  input vrst, vclk, vval, vnxt;
  input wrst, wclk, wena, wnxt;
  input [31:0] vdma, wdma;
  output reg vrdy, wrdy;

  input [NA-1:0] vadr,vlen;	// byte addresses
  input [MA-1:0] wadr,wlen;
  input  [ND-1:0] vdat;
  output [MD-1:0] wdat;

  wire L=0, H=1;

  wire vdmd = (NFLG&`BDMA)? H : L;
  wire vstr = (NFLG&`BSTREAM)? H : L;
  wire vnxx = vstr? vval : vnxt;
  reg  [NA:0] vcnt,vcntn;	// 2x address for counter logic
  wire [NA:0] vadrx = (vstr||vval||vnxt)? vadr:vdma;
  wire [NA:0] vcnta = vcnt+vadrx;

  wire wdmd = (MFLG&`BDMA)? H : L;
  wire wstr = (MFLG&`BSTREAM)? H : L;
  wire wnxx = wstr? wena : wnxt;
  reg  [MA:0] wcnt,wcntn;	// 2x address for counter logic
  wire [MA:0] wadrx = (wstr||wena||wnxt)? wadr:wdma;
  wire [MA:0] wcnta = wcnt+wadrx;

  wire [BM:0] idif = vcntn[NA:NA-BM]-wcnt[MA:MA-BM];
  wire [BM:0] odif = vcnt[NA:NA-BM]-wcntn[MA:MA-BM];

  reg[3:0] vrdx,wrdx;

  always @(posedge vclk) begin
    if (vrst) vcnt <= 0; else if (vnxx) vcnt <= vcnta;
    vcntn <= vcnta+vlen+(1<<(NA-BM));
    vrdx <= {!idif[BM],vrdx[3:1]};
    vrdy <= (vrdx==4'hF) && !vrst;
  end
  always @(posedge wclk) begin
    if (wrst) wcnt <= 0; else if (wnxx) wcnt <= wcnta;
    wcntn <= wcnta+wlen+(1<<(MA-BM));
    wrdx <= {!odif[BM],wrdx[3:1]};
    wrdy <= (wrdx==4'hF) && !wrst;
  end

  wire [ND-1:0] vdum;
  wire [MD-1:0] wdum=0;
  
  wire [NA-1:0] vaddr = vstr? vcnt : vcnta;
  wire [MA-1:0] waddr = wstr? wcnt : wcnta;
  wire [NB-1:0] vena = vval? ~0 : vdma[32-1:32-NB];

  dpram #(NKR,ND,MD,NFLG|`WONLY,MFLG|`RONLY) dps (
      vclk, H, vaddr[NA-1:LND], vena, vdat, vdum,
      wclk, H, waddr[MA-1:LMD], L,    wdum, wdat);

endmodule

module fifoNxM (iclk,irst, irdy,ival,idat, 
                oclk,orst, ordy,oena,odat);
  parameter ND=8;
  parameter MD=8;
  parameter DP=0;
  parameter BM=3;
  parameter NFLG=0;
  parameter MFLG=0;

  input irst, iclk, ival;
  input orst, oclk, oena;
  output irdy, ordy;

  input [ND-1:0] idat;
  output [MD-1:0] odat;

  wire L=0, H=1;
  wire [ND-1:0] idum;
  wire [MD-1:0] odum = 0;
  wire idumrdy,odumrdy;

  // backwards for Virtex6+ read/write optimizations
  bififoNxM #(MD,ND,DP,BM,MFLG|`RONLY,NFLG|`WONLY) bif (
      oclk,orst, odumrdy,L,odum, ordy,oena,odat, 
      iclk,irst, irdy,ival,idat, idumrdy,L,idum, L);

endmodule


// bidirectional fifo (dir=1 vi-to-wo) (dir=0 wi-to-vo)
module bififoNxM (vclk,vrst, virdy,vival,vidat, vordy,voena,vodat, 
                  wclk,wrst, wirdy,wival,widat, wordy,woena,wodat, bdir);
  parameter ND=8;	// data width in bits
  parameter MD=8;	// data width in bits
  parameter DP=0;	// depth of 2K rams
  parameter BM=4;	// address bits in comparison 
  parameter NFLG=0;	// N flags
  parameter MFLG=0;	// M flags
  parameter DIR=0;

  localparam NR=(DP>0)? DP : (ND>288 || MD>288)? 16 : (ND>144 || MD>144)? 8 : (ND>72 || MD>72)? 4 : (ND>36 || MD>36)? 2 : 1; // num 2K rams
  localparam LRS=(NR>64)?  21 : (NR>32)?  20 : (NR>16)?  19 : (NR>8)?  18 : (NR>4)?  17 : (NR>2)?  16 : (NR>1)? 15 : 14; // log2 of ram size
  localparam LND=(ND>=512)? 9 : (ND>=256)? 8 : (ND>=128)? 7 : (ND>=64)? 6 : (ND>=32)? 5 : (ND>=16)? 4 : (ND>=8)? 3 : (ND>=4)? 2 : (ND>=2)? 1 : 0; 
  localparam LMD=(MD>=512)? 9 : (MD>=256)? 8 : (MD>=128)? 7 : (MD>=64)? 6 : (MD>=32)? 5 : (MD>=16)? 4 : (MD>=8)? 3 : (MD>=4)? 2 : (MD>=2)? 1 : 0;
  localparam NA=LRS-LND;	// LND|LMD is log2 or port widths
  localparam MA=LRS-LMD;
  localparam NAC=(NFLG&`PACKET)? LRS-`LPKTS : BM;
  localparam MAC=(MFLG&`PACKET)? LRS-`LPKTS : BM;

  input vrst, vclk, vival, voena;
  input wrst, wclk, woena, wival;
  output virdy, wordy, vordy, wirdy;

  input [ND-1:0] vidat;
  output [ND-1:0] vodat;
  input [MD-1:0] widat;
  output [MD-1:0] wodat;

  input bdir;

  wire dir = (DIR==1)? 1'b1 : bdir;

  wire vena,viena;
  wire wena,wiena;
  wire [NA-1:0] vcnt;
  wire [MA-1:0] wcnt;

  fifoaddr #(NA,MA,NAC,NFLG) vfa ( vclk,vrst, virdy,vival, vordy,voena, vena,viena,vcnt,wcnt, dir);
  fifoaddr #(MA,NA,MAC,MFLG) wfa ( wclk,wrst, wirdy,wival, wordy,woena, wena,wiena,wcnt,vcnt, !dir);

  dpram #(NR*2,ND,MD,NFLG,MFLG) dps (
      vclk, vena, vcnt, viena, vidat, vodat,
      wclk, wena, wcnt, wiena, widat, wodat);

endmodule

// fifo address generator
module fifoaddr (vclk,vrst, virdy,vival, vordy,voena, vena,viena,vcnt,wcnt, dir);

  parameter NA=8;
  parameter MA=8;
  parameter BM=4;
  parameter NFLG=0;
  localparam BMAX=(1<<BM);
  localparam AFN=(NFLG&`DBLBUF)? BMAX/2+1 : (NFLG&`TIGHT)? 1 : (NFLG&`LOOSE)? 3 : 2;
  localparam NAC=BM;

  input vclk,vrst;
  input vival,voena;
  output vena,viena;
  output virdy,vordy;
  output reg [NA-1:0] vcnt;
  input [MA-1:0] wcnt;
  input dir;

  wire L=0, H=1;
  reg [NAC-1:0] wtop,vdif;
  reg virdy,vordy,voend,wstable;
  wire [NAC-1:0] afn=(NFLG&`DBLBUF)? AFN+voend : AFN;	// increment needed blocks until read addr catches up if currently reading
  wire vval = dir? vival : voena;
  wire vclx = (NFLG&`CINV)? ~vclk : vclk;
  wire vena = (NFLG&`CENA)?  vval : H;
  wire voenc; delaypipe #(5) vd (vclx,voena,voenc);
  always @ (posedge vclx) begin
    wstable <= (NAC>=MA)? H : !wcnt[MA-NAC-1];
    if (wstable) wtop <= wcnt[MA-1:MA-NAC];
    vdif  <=  wtop  - vcnt[NA-1:NA-NAC];
    voend <= voena || voenc;
  end
  always @ (posedge vclx or posedge vrst) begin
    if (vrst) vcnt  <= 0; else if (vval) vcnt <= vcnt+1;
    if (vrst) virdy <= 0; else virdy <=  dir && ((vdif>AFN)  || (vdif==0));
    if (vrst) vordy <= 0; else vordy <= !dir && ((vdif>=afn) || (vdif>0 && !voend && (NFLG&`PACKET)));  
  end
  wire viena = dir && vival;

endmodule

// dualchannel fifo 
module dcfifoNxM (vclk,vrst, virdy,vival,vidat, 
                  wclk,wrst, wordy,woena,wodat);
  parameter ND=8;	// data width in bits
  parameter MD=8;	// data width in bits
  parameter DP=0;	// depth of 2K rams
  parameter BM=4;	// address bits in comparison 
  parameter NFLG=0;	// N flags
  parameter MFLG=0;	// M flags

  localparam NR=(DP>0)? DP : (ND>288 || MD>288)? 16 : (ND>144 || MD>144)? 8 : (ND>72 || MD>72)? 4 : (ND>36 || MD>36)? 2 : 1; // num 2K rams
  localparam LRS=(NR>64)?  21 : (NR>32)? 20 : (NR>16)? 19 : (NR>8)?  18 : (NR>4)? 17 : (NR>2)? 16 : (NR>1)? 15 : 14; // log2 of ram size
  localparam LND=(ND>=128)? 7 : (ND>=64)? 6 : (ND>=32)? 5 : (ND>=16)? 4 : (ND>=8)? 3 : (ND>=4)? 2 : (ND>=2)? 1 : 0; // log2 port width
  localparam LMD=(MD>=128)? 7 : (MD>=64)? 6 : (MD>=32)? 5 : (MD>=16)? 4 : (MD>=8)? 3 : (MD>=4)? 2 : (MD>=2)? 1 : 0;
  localparam NA=LRS-LND;
  localparam MA=LRS-LMD;
  localparam NAC=(NFLG&`PACKET)? LRS-`LPKTS : BM;
  localparam MAC=(MFLG&`PACKET)? LRS-`LPKTS : BM;

  input vrst, vclk;
  input [1:0] vival;
  output [1:0] virdy;
  input [ND-1:0] vidat;

  input wrst, wclk;
  input [2:0] woena;	// woena[2] is peek at port=2 data
  output [1:0] wordy;
  output [MD-1:0] wodat;

  wire L=0, H=1;

  wire [1:0] vordy, wirdy;
  wire [1:0] voena=0;
  wire [1:0] wival=0;

  wire [ND-1:0] vodat;
  wire [MD-1:0] widat;

  wire vena,viena,vena0,viena0,vena1,viena1;
  wire wena,wiena,wena0,wiena0,wena1,wiena1;
  wire [NA-1:0] vcnt,vcnt0,vcnt1;
  wire [MA-1:0] wcnt,wcnt0,wcnt1;

  fifoaddr #(NA-1,MA-1,NAC-1,NFLG) vfa0 ( vclk,vrst, virdy[0],vival[0], vordy[0],voena[0], vena0,viena0,vcnt0,wcnt0, H);
  fifoaddr #(MA-1,NA-1,MAC-1,MFLG) wfa0 ( wclk,wrst, wirdy[0],wival[0], wordy[0],woena[0], wena0,wiena0,wcnt0,vcnt0, L);

  fifoaddr #(NA-1,MA-1,NAC-1,NFLG) vfa1 ( vclk,vrst, virdy[1],vival[1], vordy[1],voena[1], vena1,viena1,vcnt1,wcnt1, H);
  fifoaddr #(MA-1,NA-1,MAC-1,MFLG) wfa1 ( wclk,wrst, wirdy[1],wival[1], wordy[1],woena[1], wena1,wiena1,wcnt1,vcnt1, L);

  assign vena  = vena0  || vena1;
  assign viena = viena0 || viena1;
  assign vcnt  = vival[1]? {H,vcnt1[NA-2:0]} : {L,vcnt0[NA-2:0]};

  assign wena  = wena0  || wena1;
  assign wiena = wiena0 || wiena1;
  assign wcnt  = (woena[1]|woena[2])? {H,wcnt1[MA-2:0]} : {L,wcnt0[MA-2:0]}; 

  dpram #(NR*2,ND,MD,NFLG|`WONLY,MFLG|`RONLY) dps (
      vclk, vena, vcnt, viena, vidat, vodat,
      wclk, wena, wcnt, wiena, widat, wodat);

endmodule


// multichannel fifo 
module mcfifoNxM (vclk,vrst, virdy,vival,vidat, 
                  wclk,wrst, wordy,woena,wodat);
  parameter ND=8;	// data width in bits
  parameter MD=8;	// data width in bits
  parameter DP=0;	// depth of 2K rams
  parameter BM=4;	// address bits in comparison 
  parameter NFLG=0;	// N flags
  parameter MFLG=0;	// M flags

  localparam NR=(DP>0)? DP : (ND>288 || MD>288)? 16 : (ND>144 || MD>144)? 8 : (ND>72 || MD>72)? 4 : (ND>36 || MD>36)? 2 : 1; // num 2K rams
  localparam LRS=(NR>64)?  21 : (NR>32)? 20 : (NR>16)? 19 : (NR>8)?  18 : (NR>4)? 17 : (NR>2)? 16 : (NR>1)? 15 : 14; // log2 of ram size
  localparam LND=(ND>=128)? 7 : (ND>=64)? 6 : (ND>=32)? 5 : (ND>=16)? 4 : (ND>=8)? 3 : (ND>=4)? 2 : (ND>=2)? 1 : 0; // log2 port width
  localparam LMD=(MD>=128)? 7 : (MD>=64)? 6 : (MD>=32)? 5 : (MD>=16)? 4 : (MD>=8)? 3 : (MD>=4)? 2 : (MD>=2)? 1 : 0;
  localparam NA=LRS-LND;
  localparam MA=LRS-LMD;
  localparam NAC=(NFLG&`PACKET)? LRS-`LPKTS : BM;
  localparam MAC=(MFLG&`PACKET)? LRS-`LPKTS : BM;

  input vrst, vclk;
  input [2:0] vival;
  output [2:0] virdy;
  input [ND-1:0] vidat;

  input wrst, wclk;
  input [2:0] woena;
  output [2:0] wordy;
  output [MD-1:0] wodat;

  wire L=0, H=1;

  wire [2:0] vordy, wirdy;
  wire [2:0] voena=0;
  wire [2:0] wival=0;

  wire [ND-1:0] vodat;
  wire [MD-1:0] widat;

  wire vena,viena,vena0,viena0,vena1,viena1,vena2,viena2;
  wire wena,wiena,wena0,wiena0,wena1,wiena1,wena2,wiena2;
  wire [NA-1:0] vcnt,vcnt0,vcnt1,vcnt2;
  wire [MA-1:0] wcnt,wcnt0,wcnt1,wcnt2;

  // for multichannels
  fifoaddr #(NA-1,MA-1,NAC-1,NFLG) vfa0 ( vclk,vrst, virdy[0],vival[0], vordy[0],voena[0], vena0,viena0,vcnt0,wcnt0, H);
  fifoaddr #(MA-1,NA-1,MAC-1,MFLG) wfa0 ( wclk,wrst, wirdy[0],wival[0], wordy[0],woena[0], wena0,wiena0,wcnt0,vcnt0, L);

  // for core1 direct
  fifoaddr #(NA-2,MA-2,NAC-2,NFLG) vfa1 ( vclk,vrst, virdy[1],vival[1], vordy[1],voena[1], vena1,viena1,vcnt1,wcnt1, H);
  fifoaddr #(MA-2,NA-2,MAC-2,MFLG) wfa1 ( wclk,wrst, wirdy[1],wival[1], wordy[1],woena[1], wena1,wiena1,wcnt1,vcnt1, L);

  // for core2 direct
  fifoaddr #(NA-2,MA-2,NAC-2,NFLG) vfa2 ( vclk,vrst, virdy[2],vival[2], vordy[2],voena[2], vena2,viena2,vcnt2,wcnt2, H);
  fifoaddr #(MA-2,NA-2,MAC-2,MFLG) wfa2 ( wclk,wrst, wirdy[2],wival[2], wordy[2],woena[2], wena2,wiena2,wcnt2,vcnt2, L);

  assign vena  = vena0  || vena1  || vena2;
  assign viena = viena0 || viena1 || viena2;
  assign vcnt  = vival[2]? {H,H,vcnt2[NA-3:0]} : vival[1]? {H,L,vcnt1[NA-3:0]} : {L,vcnt0[NA-2:0]};

  assign wena  = wena0  || wena1  || wena2;
  assign wiena = wiena0 || wiena1 || wiena2;
  assign wcnt  = woena[2]? {H,H,wcnt2[MA-3:0]} : woena[1]? {H,L,wcnt1[MA-3:0]} : {L,wcnt0[MA-2:0]}; 

  dpram #(NR*2,ND,MD,NFLG|`WONLY,MFLG|`RONLY) dps (
      vclk, vena, vcnt, viena, vidat, vodat,
      wclk, wena, wcnt, wiena, widat, wodat);

endmodule

module fifoNxMpkt2 (iclk,irst, irdy,ival,idat, 
                    oclk,orst, ordy,oena,odat, opkt_,size,osel);
  parameter ND=32;
  parameter MD=64;
  parameter DP=0;
  parameter BM=3;
  parameter NFLG=0;
  parameter MFLG=0;
  parameter NC=8;		// log2(channels)

  localparam NCHN=(1<<NC);
  localparam NPKT=`DMA_PKT_LEN;	// num 64by chunks per DMA packet
  localparam NPKTB=(NPKT==4)? 2 : (NPKT==2)? 1 : 0;
  localparam NKB=NCHN*64*2*NPKT/1024;
  localparam NP=NPKTB + ((ND==8)? 6 : (ND==16)? 5 : (ND==32)? 4 : 3);
  localparam NCP=NC+NP;
  localparam NDH=ND/2;
  localparam NKBH=NKB/2;

  input irst, iclk, ival;
  input orst, oclk, oena;
  output irdy, ordy;

  input [ND-1:0] idat;
  output [MD-1:0] odat;

  input opkt_;
  input [NC-1:0] size;
  output reg [NC-1:0] osel;

  wire L=0, H=1;
  wire [ND-1:0] jdat;
  wire [NP-1:0] full=~0;

  // input ram addressing 
  reg opkt;
  reg pass,primed,jval,kval;
  reg [NC-1:0] icnt,sizex;
  reg [NP-1:0] pcnt;
  reg [NCP-1:0] jcnt;
  reg [ND-1:0] kdat;
  wire izer = ival && (icnt==sizex);
  wire jzer = ival && (jcnt=={sizex,full});
  always @(posedge iclk) begin
    opkt <= opkt_;
    sizex <= opkt? size : 0;
    if (irst|izer) icnt <= 0; else if (ival) icnt <= icnt+1;
    if (irst|jzer) pcnt <= 0; else if (izer) pcnt <= pcnt+1;
    if (irst|jzer) jcnt <= 0; else if (ival) jcnt <= jcnt+1;
    if (irst)      pass <= 0; else if (jzer) pass <= !pass;
    if (irst)    primed <= 0; else if (jzer) primed <= H;
    jval <= opkt? ival && primed : L;
    kval <= opkt? jval : ival;
    kdat <= opkt? jdat : idat;
  end
  wire [NCP:0] iadr = {pass,icnt,pcnt};
  wire [NCP:0] jadr = {!pass,jcnt};

  // dual port RAM for corner turn
  sdpram #(NKBH,NDH,NDH) dpse (iclk,H, iadr,ival, idat[NDH-1: 0],   iclk,H, jadr,jdat[NDH-1: 0]);
  sdpram #(NKBH,NDH,NDH) dpso (iclk,H, iadr,ival, idat[ND-1:NDH],   iclk,H, jadr,jdat[ND-1:NDH]);

  // regular output fifio
  fifoNxM #(ND,MD,DP,BM,NFLG,MFLG) bif (iclk,irst, irdy,kval,kdat, oclk,orst, ordy,oena,odat);

  // output channel select
  reg [3:0] ocnt;
  wire osena = oena && ((MD==32)? (ocnt[3:0]==15) : (ocnt[2:0]==7));
  wire osrst = orst || (osena && (osel==sizex));
  wire ocrst = orst || !opkt;
  always @(posedge oclk) begin
    if (ocrst) ocnt <= 0; else if (oena)  ocnt <= ocnt+1;
    if (osrst) osel <= 0; else if (osena) osel <= osel+1;
  end

endmodule


// 2-D corner turn + NC chan x NPKTx64by packetizer + FIFO
module fifoNxMpkt (iclk,irst, irdy,ival,idat, 
                   oclk,orst, ordy,oena,odat, opkt_,oraw_,oreal_,size,osel);
  parameter ND=32;
  parameter MD=64;
  parameter DP=0;
  parameter BM=3;
  parameter NFLG=0;
  parameter MFLG=0;
  parameter NC=8;		// log2(channels)

  localparam NCHN=(1<<NC);
  localparam NPKT=`DMA_PKT_LEN;	// num 64by chunks per DMA packet
  localparam NPKTB=(NPKT==4)? 2 : (NPKT==2)? 1 : 0;
  localparam NKB=NCHN*64*2*NPKT/1024;
  localparam NP=NPKTB + ((ND==8)? 6 : (ND==16)? 5 : (ND==32)? 4 : 3);
  localparam NCP=NC+NP;
  localparam NDH=ND/2;
  localparam NKBH=NKB/2;

  input irst, iclk, ival;
  input orst, oclk, oena;
  output irdy, ordy;

  input [ND-1:0] idat;
  output [MD-1:0] odat;

  input oreal_,opkt_,oraw_;
  input [NC-1:0] size;
  output reg [NC-1:0] osel;

  wire L=0, H=1;
  wire [ND-1:0] jdat;
  wire [NP-1:0] full=~0;

  // input ram addressing 
  reg oreal,opkt,oraw;
  reg pass,primed,jval,kval,pha,iq;
  reg [NC-1:0] icnt,sizex;
  reg [NP-1:0] pcnt;
  reg [NCP-1:0] jcnt;
  reg [ND-1:0] kdat;
  wire iinc = ival && (oreal? H    : iq);
  wire izer = iinc && (icnt==sizex);
  wire jinc = ival && pha;
  wire qinc = ival && ((oreal&opkt)? izer : H );
  wire pinc = izer && iq;
  wire jzer = ival && (jcnt=={sizex,full}) && pha;
  wire ival0 = ival && (oraw || !iq);
  wire ival1 = ival && (oraw ||  iq);
  always @(posedge iclk) begin
    oraw <= oraw_;
    opkt <= opkt_;
    oreal <= oreal_;
    sizex <= opkt? size : 0;
    if (irst) pha <= 0; else if (oraw) pha <= 1; else if (ival) pha  <= !pha;
    if (irst)  iq <= 0; else if (oraw)  iq <= 1; else if (qinc) iq   <= !iq;
    if (irst|izer) icnt <= 0; else if (iinc) icnt <= icnt+1;
    if (irst|jzer) pcnt <= 0; else if (pinc) pcnt <= pcnt+1;
    if (irst|jzer) jcnt <= 0; else if (jinc) jcnt <= jcnt+1;
    if (irst)    primed <= 0; else if (jzer) primed <= H;
    if (irst)      pass <= 0; else if (jzer) pass <= !pass;
    jval <= jinc && primed;
    kval <= jval;
    kdat <= jdat;
//if (ival) $write("iinc=%d iq=%d pha=%d icnt=%x izer=%d pcnt=%x jcnt=%x jzer=%d\n",iinc,iq,pha,icnt,izer,pcnt,jcnt,jzer);
  end
  wire [NCP:0] iadr = {pass,icnt,pcnt};
  wire [NCP:0] jadr = {!pass,jcnt};

  // dual port RAM for corner turn
  sdpram #(NKBH,NDH,NDH) dpse (iclk,H, iadr,ival0, idat[NDH-1: 0],   iclk,H, jadr,jdat[NDH-1: 0]);
  sdpram #(NKBH,NDH,NDH) dpso (iclk,H, iadr,ival1, idat[ND-1:NDH],   iclk,H, jadr,jdat[ND-1:NDH]);

  // regular output fifio
  fifoNxM #(ND,MD,DP,BM,NFLG,MFLG) bif (iclk,irst, irdy,kval,kdat, oclk,orst, ordy,oena,odat);

  // output channel select
  reg [3:0] ocnt;
  wire osena = oena && ((MD==32)? (ocnt[3:0]==15) : (ocnt[2:0]==7));
  wire osrst = orst || (osena && (osel==sizex));
  always @(posedge oclk) begin
    if (orst)  ocnt <= 0; else if (oena) ocnt <= ocnt+1;
    if (osrst) osel <= 0; else if (osena) osel <= osel+1;
  end

endmodule

// single port RAM 
module spram (vclk, vena, vaddr, viena, vidat, vodat);

  parameter NKB=2;	// kbytes of ram desired
  parameter NW=8;	// width for port A in bits
  parameter NFLG=0;	// flags for port A

  localparam MW=NW;
  localparam LA=24;	// enough address bits to pass through to dpram without computing here

  input vclk,vena,viena;
  input [LA-1:0] vaddr;
  input [NW-1:0] vidat;
  output [NW-1:0] vodat;

  wire wclk=vclk;
  wire wena=1'b0;
  wire wiena=1'b0;
  wire [LA-1:0] waddr;
  wire [MW-1:0] wodat;
  wire [MW-1:0] widat;	// dummy bus

  dpram #(NKB,NW,MW,NFLG,0) dpr 
    (vclk, vena, vaddr, viena, vidat, vodat,
     wclk, wena, waddr, wiena, widat, wodat);

endmodule 

// simple dual port RAM with WONLY/RONLY port restrictions
module sdpram (vclk,vena, vaddr, viena, vidat, 
               wclk,wena, waddr,        wodat);

  parameter NKB=2;	// kbytes of ram desired
  parameter NW=8;	// width for port A in bits
  parameter MW=NW;	// width for port B in bits
  parameter NFLG=0;	// flags for port A
  parameter MFLG=0;	// flags for port B

  localparam LA=24;	// enough address bits to pass through to dpram without computing here

  input vclk,vena,viena;
  input [LA-1:0] vaddr;
  input [NW-1:0] vidat;

  input wclk,wena;
  input [LA-1:0] waddr;
  output [MW-1:0] wodat;

  wire [NW-1:0] vodat;	// dummy bus
  wire  [MW-1:0] widat;	// dummy bus
  wire wiena=1'b0;

  dpram #(NKB,MW,NW,MFLG|`RONLY,NFLG|`WONLY) dpr 
    (wclk, wena, waddr, wiena, widat, wodat,
     vclk, vena, vaddr, viena, vidat, vodat);

endmodule

// dual port RAM with full R/W from both ports
module dpram (vclk, vena, vaddr, viena, vidat, vodat,
              wclk, wena, waddr, wiena, widat, wodat);

  parameter signed NKB=2;	// kbytes of ram desired or -NR actual rams
  parameter NW=8;		// width for port A in bits
  parameter MW=NW;		// width for port B in bits
  parameter NFLG=0;		// flags for port A
  parameter MFLG=0;		// flags for port B

  localparam LW=(NW<MW)? NW:MW;		// lesser of port widths
  localparam GW=(NW<MW)? MW:NW;		// greater of port widths
  localparam LWB=(LW>=512)? 64 : (LW>=256)? 32 : (LW>=128)? 16 : (LW>=64)? 8 : LW/8;	// lesser of port width in bytes
  localparam GWB=(GW>=512)? 64 : (GW>=256)? 32 : (GW>=128)? 16 : (GW>=64)? 8 : GW/8;	// greater of port width in bytes

  localparam BPAR=((NFLG&`BPAR) || (MFLG&`BPAR))? 1:0;			// use parity in block mode at top of word
  localparam SDPP=((NFLG&`RONLY) && (MFLG&`WONLY) && `RAMKB==4)? 1:0;	// possible simple dual port mode for Virtex6+
  localparam F2K=(NKB<=2 && GW<=(SDPP?36:18))? 1:0;			// fits in a 2K RAM18 block
  localparam SDP=(SDPP==0)? 0 : (F2K!=0)? 4 : 8;			// simple dual port mode

  localparam RKB=(`RAMKB==4 && F2K!=0)? 2 : `RAMKB;		// dont use 32K ram if 16K will do
  localparam NR=(NKB<0)? -NKB : (NKB<RKB)? 1 : NKB/RKB;		// # of total rams needed by size
  localparam PAR=(LW>8 && LW==LWB*9 && NR<=LWB)?1:0;		// data ports use parity bit
  localparam BPBL=(8+PAR);					// num usable bits per byte lane
  localparam MRW=(((SDP>0)? SDP*8:`RAMBW)/8)*BPBL;		// max supported ram port width
  localparam NRV=(GW+MRW-1)/MRW;				// min rams to support width
  localparam NRX=(LW<NR)? LW : (NRV>NR)? NRV : NR;		// # rams wide to use
  localparam NRW=(PAR!=0 && NRX>LWB)? LWB : NRX;		// # rams wide if using parity bits
  localparam NRD=(NR<NRW)? 1 : NR/NRW;				// # rams deep - has to mux output of NRD rams
  localparam LRD=(NRD>8)? 4 : (NRD>4)? 3 : (NRD>2)? 2 : (NRD>1)? 1 : 0;	// log2 of rams deep

  localparam ND=NW/NRW;			// bit width per ram
  localparam MD=MW/NRW;
  localparam LD=(ND<MD)? ND:MD;		// lesser of port widths per ram
  localparam NB=(ND>36)? 8 : (ND>18)? 4 : (ND>9)? 2 : (ND/8);	// data bytes per ram
  localparam MB=(MD>36)? 8 : (MD>18)? 4 : (MD>9)? 2 : (MD/8);	// data bytes per ram
  localparam LND=(ND>=64)? 6 : (ND>=32)? 5 : (ND>=16)? 4 : (ND>=8)? 3 : (ND>=4)? 2 : (ND>=2)? 1 : 0; // log2 port width per ram
  localparam LMD=(MD>=64)? 6 : (MD>=32)? 5 : (MD>=16)? 4 : (MD>=8)? 3 : (MD>=4)? 2 : (MD>=2)? 1 : 0;
  localparam LRS=(RKB==4)? 15 : (RKB==2)? 14 : 13;	// log2 ram size
  localparam NA=LRS-LND+LRD;
  localparam MA=LRS-LMD+LRD;
  localparam NV=(NB==0)? ND:NB*8;	// padded data bits
  localparam MV=(MB==0)? MD:MB*8;
  localparam NP=(NB==0)? 1:NB;		// parity bits
  localparam MP=(MB==0)? 1:MB;
  localparam LS=(NB!=0 && MB!=0)? 8 : LD;
  localparam NX=(ND>NV && BPAR==0)? 9 : LS;
  localparam MX=(MD>MV && BPAR==0)? 9 : LS;
  localparam NIR=(NFLG&`IREG)?1:0;	// input register to aid timing - particularly with NRD>1
  localparam MIR=(MFLG&`IREG)?1:0;	// input register to aid timing - particularly with NRD>1
  localparam NOR=(NFLG&`OREG)?1:0;	// output register to aid timing - particularly with NRD>1
  localparam MOR=(MFLG&`OREG)?1:0;	// output register to aid timing - particularly with NRD>1
  localparam NWX=NRW*(NV+NP);
  localparam MWX=NRW*(MV+MP);
  localparam NBE=(NFLG&`BENA)? NW/8:1;
  localparam MBE=(MFLG&`BENA)? MW/8:1;
  localparam NFLGY=((NFLG&`BENA)!=0 && NB<=1)? NFLG^`BENA : NFLG;
  localparam MFLGY=((MFLG&`BENA)!=0 && MB<=1)? MFLG^`BENA : MFLG;
  localparam NFLGX=(NOR>0 && LRD>0)? NFLGY^`OREG : NFLGY;
  localparam MFLGX=(MOR>0 && LRD>0)? MFLGY^`OREG : MFLGY;
  
  input vclk,vena;
  input [NBE-1:0] viena;
  input [NA-1:0] vaddr;
  input [NWX-1:0] vidat;
  output [NWX-1:0] vodat;

  input wclk,wena;
  input [MBE-1:0] wiena;
  input [MA-1:0] waddr;
  input [MWX-1:0] widat;
  output [MWX-1:0] wodat;

  // handle extra input register
  wire [NA-1:0]  vaddx; ff #(NA, NIR) ffn1 (vaddx,vaddr,vclk);
  wire [NWX-1:0] vidax; ff #(NWX,NIR) ffn2 (vidax,vidat,vclk);
  wire [NBE-1:0] vienx; ff #(NBE,NIR) ffn3 (vienx,viena,vclk);
  wire [MA-1:0]  waddx; ff #(MA, MIR) ffm1 (waddx,waddr,wclk);
  wire [MWX-1:0] widax; ff #(MWX,MIR) ffm2 (widax,widat,wclk);
  wire [MBE-1:0] wienx; ff #(MBE,MIR) ffm3 (wienx,wiena,wclk);

  genvar i,j;
  generate
  for (j=0; j<NRW; j=j+1) begin:dpw
    // split rams by width
    wire [NV-1:0] vid,vod;
    wire [NP-1:0] vip,vop;
    wire [MV-1:0] wid,wod;
    wire [MP-1:0] wip,wop;
    for (i=0; i<NV/LS; i=i+1) begin:ndc
      assign vid[LS-1+i*LS:i*LS] = vidax[LS-1+(j+i*NRW)*NX:(j+i*NRW)*NX];
      assign vodat[LS-1+(j+i*NRW)*NX:(j+i*NRW)*NX] = vod[LS-1+i*LS:i*LS];
    end
    for (i=0; i<MV/LS; i=i+1) begin:mdc
      assign wid[LS-1+i*LS:i*LS] = widax[LS-1+(j+i*NRW)*MX:(j+i*NRW)*MX];
      assign wodat[LS-1+(j+i*NRW)*MX:(j+i*NRW)*MX] = wod[LS-1+i*LS:i*LS];
    end
    if (ND>NV) for (i=0; i<NB; i=i+1) begin:npc
      assign vip[i] = vidax[BPAR? j+NRW*(i+NV) : (j+i*NRW+1)*NX-1];
      assign vodat[BPAR? j+NRW*(i+NV) : (j+i*NRW+1)*NX-1] = vop[i];
    end
    if (MD>MV) for (i=0; i<MB; i=i+1) begin:mpc
      assign wip[i] = widax[BPAR? j+NRW*(i+MV) : (j+i*NRW+1)*MX-1];
      assign wodat[BPAR? j+NRW*(i+MV) : (j+i*NRW+1)*MX-1] = wop[i];
    end
    // handle write byte enables - unary version has issues
    wire [NP-1:0] vjenx;
    wire [MP-1:0] wjenx;
    for (i=0; i<NP; i=i+1) begin:nbe
      if (NFLG&`BENA) assign vjenx[i] = vienx[((j+i*NRW)*NX)/8]; else assign vjenx[i] = vienx;
    end
    for (i=0; i<MP; i=i+1) begin:mbe
      if (MFLG&`BENA) assign wjenx[i] = wienx[((j+i*NRW)*MX)/8]; else assign wjenx[i] = wienx;
    end
    // split rams by depth
    if (LRD>0) begin:mdr
      wire [NRD*NV-1:0] vodx;
      wire [NRD*MV-1:0] wodx;
      wire [NRD*NP-1:0] vopx;
      wire [NRD*MP-1:0] wopx;
      for (i=0; i<NRD; i=i+1) begin:dpd
        wire [NP-1:0] vjenxi = (vaddx[NA-1:NA-LRD]==i)?vjenx:0;
        wire [MP-1:0] wjenxi = (waddx[MA-1:MA-LRD]==i)?wjenx:0;
        dpramb #(RKB,NV,MV,NFLGX,MFLGX) inst (
          vclk,vena, vaddx[NA-LRD-1:0], vjenxi, vid,vip, vodx[NV-1+i*NV:i*NV],vopx[NP-1+i*NP:i*NP],
          wclk,wena, waddx[MA-LRD-1:0], wjenxi, wid,wip, wodx[MV-1+i*MV:i*MV],wopx[MP-1+i*MP:i*MP]);
      end
      reg [LRD-1:0] vsel; always @(posedge vclk) if (vena) vsel <= vaddx[NA-1:NA-LRD];
      reg [LRD-1:0] wsel; always @(posedge wclk) if (wena) wsel <= waddx[MA-1:MA-LRD];
      muxMxNp #(NRD,NV,NOR) mxnd (vod,vodx,vsel,vclk);
      muxMxNp #(NRD,NP,NOR) mxnp (vop,vopx,vsel,vclk);
      muxMxNp #(NRD,MV,MOR) mxmd (wod,wodx,wsel,wclk);
      muxMxNp #(NRD,MP,MOR) mxmp (wop,wopx,wsel,wclk);
    end
    if (LRD<=0) begin:sdr
      dpramb #(RKB,NV,MV,NFLGX,MFLGX,(SDP>0)?1:0) inst (
        vclk,vena, vaddx,vjenx, vid,vip, vod,vop,
        wclk,wena, waddx,wjenx, wid,wip, wod,wop);
    end
  end
  endgenerate

endmodule

module dpramb (vclk, vena, vaddr, viena, vid,vip, vod,vop,
               wclk, wena, waddr, wiena, wid,wip, wod,wop);

  parameter RKB=2;	// RAM depth in KB
  parameter ND=8;	// port A width
  parameter MD=8;	// port B width
  parameter NFLG=0;	// port A flags
  parameter MFLG=0;	// port B flags
  parameter SDP=0;

  localparam LRS=(RKB>4)? 16 : (RKB>2)? 15 : (RKB>1)? 14 : 13;	// log2 ram size in bits

  localparam NB=(ND>36)? 8 : (ND>18)? 4 : (ND>9)? 2 : (ND>4)? 1 : 0;
  localparam MB=(MD>36)? 8 : (MD>18)? 4 : (MD>9)? 2 : (MD>4)? 1 : 0;
  localparam LND=(NB==8)? 6 : (NB==4)? 5 : (NB==2)? 4 : (NB==1)? 3 : (ND==4)? 2 : (ND==2)? 1 : 0; // log2 port width
  localparam LMD=(MB==8)? 6 : (MB==4)? 5 : (MB==2)? 4 : (MB==1)? 3 : (MD==4)? 2 : (MD==2)? 1 : 0; // log2 port width
  localparam NA=LRS-LND;
  localparam MA=LRS-LMD;
  localparam NW=(NB==0)?ND:NB*9;
  localparam MW=(MB==0)?MD:MB*9;
  localparam NV=(NB==0)?ND:NB*8;
  localparam MV=(MB==0)?MD:MB*8;
  localparam NP=(NB==0)?1:NB;
  localparam MP=(MB==0)?1:MB;
  localparam GW=(MW>NW)?MW:NW;

  localparam NREG=(NFLG&`OREG)?1:0;
  localparam MREG=SDP?NREG:(MFLG&`OREG)?1:0;
  localparam NCI=(NFLG&`CINV)?1:0;
  localparam MCI=(MFLG&`CINV)?1:0;
  localparam NEI=(NFLG&`EINV)?1:0;
  localparam MEI=(MFLG&`EINV)?1:0;
  localparam NWM=(NFLG&`RFIRST)?"READ_FIRST":(NFLG&`WFIRST)?"WRITE_FIRST":SDP?"WRITE_FIRST":"NO_CHANGE";
  localparam MWM=(MFLG&`RFIRST)?"READ_FIRST":(MFLG&`WFIRST)?"WRITE_FIRST":SDP?"WRITE_FIRST":"NO_CHANGE";
  
  input vclk,vena;
  input [NP-1:0] viena;
  input [NA-1:0] vaddr;
  input [ND-1:0] vid;
  input [NP-1:0] vip;
  output [ND-1:0] vod;
  output [NP-1:0] vop;

  input wclk,wena;
  input [MP-1:0] wiena;
  input [MA-1:0] waddr;
  input [MD-1:0] wid;
  input [MP-1:0] wip;
  output [MD-1:0] wod;
  output [MP-1:0] wop;

  wire H=1, L=0;

  wire [7:0] vjena = (NFLG&`BENA)? {8{viena}} : {8{viena[0]}};
  wire [7:0] wjena = (MFLG&`BENA)? {8{wiena}} : {8{wiena[0]}};

`ifdef NORAMREG

  wire [ND-1:0] vodb;
  wire [NP-1:0] vopb;
  reg [ND-1:0]  vodr; always @(posedge vclk) vodr <= vodb;
  reg [NP-1:0]  vopr; always @(posedge vclk) vopr <= vopb;

  wire [MD-1:0] wodb;
  wire [MP-1:0] wopb;
  reg [MD-1:0]  wodr; always @(posedge wclk) wodr <= wodb;
  reg [MP-1:0]  wopr; always @(posedge wclk) wopr <= wopb;

  assign vod = NREG? vodr : vodb;
  assign vop = NREG? vopr : vopb;
  assign wod = MREG? wodr : wodb;
  assign wop = MREG? wopr : wopb;

`endif

`ifdef USERAMBXX

  generate
  if (MW>NW) begin:ab
    RAMBXX #(NV,MV,LRS,NFLG,MFLG) ram (
    .CLKA(NCI?~vclk:vclk),.ENA(vena),.WEA_(vjena),.SSRA(L), .ADDRA(vaddr), .DIA_(vid),.DIPA_(vip), .DOA(vod),.DOPA(vop),
    .CLKB(MCI?~wclk:wclk),.ENB(wena),.WEB_(wjena),.SSRB(L), .ADDRB(waddr), .DIB_(wid),.DIPB_(wip), .DOB(wod),.DOPB(wop));
  end
  else begin:ba
    RAMBXX #(MV,NV,LRS,MFLG,NFLG) ram (
    .CLKB(NCI?~vclk:vclk),.ENB(vena),.WEB_(vjena),.SSRB(L), .ADDRB(vaddr), .DIB_(vid),.DIPB_(vip), .DOB(vod),.DOPB(vop),
    .CLKA(MCI?~wclk:wclk),.ENA(wena),.WEA_(wjena),.SSRA(L), .ADDRA(waddr), .DIA_(wid),.DIPA_(wip), .DOA(wod),.DOPA(wop));
  end
  endgenerate

`elsif ALTERA

  localparam RBS=((NW<9)||(MW<9))?8:10;
  localparam ANV=(RBS==8)?NV:NB*10;
  localparam AMV=(RBS==8)?MV:MB*10;
    
  wire [ANV-1:0] vidp,vodp;
  wire [AMV-1:0] widp,wodp;  
  wire [NB-1:0] dumn;
  wire [MB-1:0] dumm;
  
  genvar i;  
  generate
  if (RBS==8) begin
    assign vidp = vid;
    assign vod  = vodp;
    assign widp = wid;
    assign wod  = wodp;	 	 
  end
  else begin	 
    for(i=0; i<NB; i=i+1) begin: ALT_APARITY
      assign vidp[(i+1)*10-1:i*10] = {1'b0,vip[i],vid[(i+1)*8-1:i*8]};
      assign {dumn[i],vop[i],vod[(i+1)*8-1:i*8]} = vodp[(i+1)*10-1:i*10];
    end
    for(i=0; i<MB; i=i+1) begin: ALT_BPARITY      
      assign widp[(i+1)*10-1:i*10] = {1'b0,wip[i],wid[(i+1)*8-1:i*8]};
      assign {dumm[i],wop[i],wod[(i+1)*8-1:i*8]} = wodp[(i+1)*10-1:i*10];
    end
  end
  endgenerate

  altera_syncram #(
    .lpm_type("altsyncram"), .intended_device_family("Stratix V"),
    .operation_mode("BIDIR_DUAL_PORT"), .byte_size(RBS), .ram_block_type("M20K"),	// M9K,M144K,AUTO
    .width_a(ANV), .widthad_a(NA), .numwords_a(1<<NA), .width_byteena_a(1), .outdata_reg_a(NREG? "CLOCK0": "UNREGISTERED"),
    .width_b(AMV), .widthad_b(MA), .numwords_b(1<<MA), .width_byteena_b(1), .outdata_reg_b(MREG? "CLOCK1": "UNREGISTERED"),
    .clock_enable_input_a("BYPASS"), .clock_enable_output_a("BYPASS"), .read_during_write_mode_port_a("NEW_DATA_NO_NBE_READ"),
    .clock_enable_input_b("BYPASS"), .clock_enable_output_b("BYPASS"), .read_during_write_mode_port_b("NEW_DATA_NO_NBE_READ"),
    .indata_reg_b("CLOCK1"), .address_reg_b("CLOCK1"), .byteena_reg_b("CLOCK1"), .wrcontrol_wraddress_reg_b("CLOCK1"),
    .outdata_aclr_a("NONE"), .outdata_aclr_b("NONE"),
    .power_up_uninitialized("FALSE")
  ) b9k (
    .clock0 (vclk), .aclr0 (L), .rden_a (H), .byteena_a (H), .wren_a (viena), .address_a (vaddr), .data_a (vidp), .q_a (vodp),
    .clock1 (wclk), .aclr1 (L), .rden_b (H), .byteena_b (H), .wren_b (wiena), .address_b (waddr), .data_b (widp), .q_b (wodp),
    .addressstall_a (L), .addressstall_b (L),
    .clocken0 (vena), .clocken1 (wena),
    .clocken2 (vena), .clocken3 (wena),
    .eccstatus ()
  );
  
`elsif VIRTEX2

  generate 
  if (NW==1 && MW==1) begin:c1x1
    RAMB16_S1_S1 ram (
    .ADDRA(vaddr), .DIA(vid), .CLKA(NCI?~vclk:vclk), .ENA(NEI?~vena:vena), .WEA(viena), .SSRA(L), .DOA(vodb),
    .ADDRB(waddr), .DIB(wid), .CLKB(MCI?~wclk:wclk), .ENB(MEI?~wena:wena), .WEB(wiena), .SSRB(L), .DOB(wodb));
  end
  if (NW==1 && MW==2) begin:c1x2
    RAMB16_S1_S2 ram (
    .ADDRA(vaddr), .DIA(vid), .CLKA(NCI?~vclk:vclk), .ENA(NEI?~vena:vena), .WEA(viena), .SSRA(L), .DOA(vodb),
    .ADDRB(waddr), .DIB(wid), .CLKB(MCI?~wclk:wclk), .ENB(MEI?~wena:wena), .WEB(wiena), .SSRB(L), .DOB(wodb));
  end
  if (NW==2 && MW==1) begin:c2x1
    RAMB16_S1_S2 ram (
    .ADDRA(waddr), .DIA(wid), .CLKA(NCI?~wclk:wclk), .ENA(NEI?~wena:wena), .WEA(wiena), .SSRA(L), .DOA(wodb),
    .ADDRB(vaddr), .DIB(vid), .CLKB(MCI?~vclk:vclk), .ENB(MEI?~vena:vena), .WEB(viena), .SSRB(L), .DOB(vodb));
  end
  if (NW==2 && MW==2) begin:c2x2
    RAMB16_S2_S2 ram (
    .ADDRA(vaddr), .DIA(vid), .CLKA(NCI?~vclk:vclk), .ENA(NEI?~vena:vena), .WEA(viena), .SSRA(L), .DOA(vodb),
    .ADDRB(waddr), .DIB(wid), .CLKB(MCI?~wclk:wclk), .ENB(MEI?~wena:wena), .WEB(wiena), .SSRB(L), .DOB(wodb));
  end
  if (NW==4 && MW==4) begin:c4x4
    RAMB16_S4_S4 ram (
    .ADDRA(vaddr), .DIA(vid), .CLKA(NCI?~vclk:vclk), .ENA(NEI?~vena:vena), .WEA(viena), .SSRA(L), .DOA(vodb),
    .ADDRB(waddr), .DIB(wid), .CLKB(MCI?~wclk:wclk), .ENB(MEI?~wena:wena), .WEB(wiena), .SSRB(L), .DOB(wodb));
  end
  if (NW==4 && MW==9) begin:c4x9
    RAMB16_S4_S9 ram (
    .ADDRA(vaddr), .DIA(vid), .CLKA(NCI?~vclk:vclk), .ENA(NEI?~vena:vena), .WEA(viena), .SSRA(L), .DOA(vodb),
    .ADDRB(waddr), .DIB(wid), .CLKB(MCI?~wclk:wclk), .ENB(MEI?~wena:wena), .WEB(wiena), .SSRB(L), .DOB(wodb),.DIPB(wip));
  end
  if (NW==4 && MW==18) begin:c4x18
    RAMB16_S4_S18 ram (
    .ADDRA(vaddr), .DIA(vid), .CLKA(NCI?~vclk:vclk), .ENA(NEI?~vena:vena), .WEA(viena), .SSRA(L), .DOA(vodb),
    .ADDRB(waddr), .DIB(wid), .CLKB(MCI?~wclk:wclk), .ENB(MEI?~wena:wena), .WEB(wiena), .SSRB(L), .DOB(wodb),.DIPB(wip));
  end
  if (NW==4 && MW==36) begin:c4x36
    RAMB16_S4_S36 ram (
    .ADDRA(vaddr), .DIA(vid), .CLKA(NCI?~vclk:vclk), .ENA(NEI?~vena:vena), .WEA(viena), .SSRA(L), .DOA(vodb),
    .ADDRB(waddr), .DIB(wid), .CLKB(MCI?~wclk:wclk), .ENB(MEI?~wena:wena), .WEB(wiena), .SSRB(L), .DOB(wodb),.DIPB(wip));
  end
  if (NW==9 && MW==9) begin:c9x9
    RAMB16_S9_S9 ram (
    .ADDRA(vaddr), .DIA(vid),.DIPA(vip), .CLKA(NCI?~vclk:vclk), .ENA(NEI?~vena:vena), .WEA(viena), .SSRA(L), .DOA(vodb),.DOPA(vopb),
    .ADDRB(waddr), .DIB(wid),.DIPB(wip), .CLKB(MCI?~wclk:wclk), .ENB(MEI?~wena:wena), .WEB(wiena), .SSRB(L), .DOB(wodb),.DOPB(wopb));
  end
  if (NW==9 && MW==18) begin:c9x18
    RAMB16_S9_S18 ram (
    .ADDRA(vaddr), .DIA(vid),.DIPA(vip), .CLKA(NCI?~vclk:vclk), .ENA(NEI?~vena:vena), .WEA(viena), .SSRA(L), .DOA(vodb),.DOPA(vopb),
    .ADDRB(waddr), .DIB(wid),.DIPB(wip), .CLKB(MCI?~wclk:wclk), .ENB(MEI?~wena:wena), .WEB(wiena), .SSRB(L), .DOB(wodb),.DOPB(wopb));
  end
  if (NW==9 && MW==36) begin:c9x36
    RAMB16_S9_S36 ram (
    .ADDRA(vaddr), .DIA(vid),.DIPA(vip), .CLKA(NCI?~vclk:vclk), .ENA(NEI?~vena:vena), .WEA(viena), .SSRA(L), .DOA(vodb),.DOPA(vopb),
    .ADDRB(waddr), .DIB(wid),.DIPB(wip), .CLKB(MCI?~wclk:wclk), .ENB(MEI?~wena:wena), .WEB(wiena), .SSRB(L), .DOB(wodb),.DOPB(wopb));
  end
  if (NW==18 && MW==9) begin:c18x9
    RAMB16_S9_S18 ram (
    .ADDRA(waddr), .DIA(wid),.DIPA(wip), .CLKA(MCI?~wclk:wclk), .ENA(MEI?~wena:wena), .WEA(wiena), .SSRA(L), .DOA(wodb),.DOPA(wopb),
    .ADDRB(vaddr), .DIB(vid),.DIPB(vip), .CLKB(NCI?~vclk:vclk), .ENB(NEI?~vena:vena), .WEB(viena), .SSRB(L), .DOB(vodb),.DOPB(vopb));
  end
  if (NW==18 && MW==18) begin:c18x18
    RAMB16_S18_S18 ram (
    .ADDRA(vaddr), .DIA(vid),.DIPA(vip), .CLKA(NCI?~vclk:vclk), .ENA(NEI?~vena:vena), .WEA(viena), .SSRA(L), .DOA(vodb),.DOPA(vopb),
    .ADDRB(waddr), .DIB(wid),.DIPB(wip), .CLKB(MCI?~wclk:wclk), .ENB(MEI?~wena:wena), .WEB(wiena), .SSRB(L), .DOB(wodb),.DOPB(wopb));
  end
  if (NW==18 && MW==36) begin:c18x36
    RAMB16_S18_S36 ram (
    .ADDRA(vaddr), .DIA(vid),.DIPA(vip), .CLKA(NCI?~vclk:vclk), .ENA(NEI?~vena:vena), .WEA(viena), .SSRA(L), .DOA(vodb),.DOPA(vopb),
    .ADDRB(waddr), .DIB(wid),.DIPB(wip), .CLKB(MCI?~wclk:wclk), .ENB(MEI?~wena:wena), .WEB(wiena), .SSRB(L), .DOB(wodb),.DOPB(wopb));
  end
  if (NW==36 && MW==4) begin:c36x4
    RAMB16_S4_S36 ram (
    .ADDRA(waddr), .DIA(wid), .CLKA(MCI?~wclk:wclk), .ENA(MEI?~wena:wena), .WEA(wiena), .SSRA(L), .DOA(wodb),
    .ADDRB(vaddr), .DIB(vid), .CLKB(NCI?~vclk:vclk), .ENB(NEI?~vena:vena), .WEB(viena), .SSRB(L), .DOB(vodb),.DOPB(vopb));
  end
  if (NW==36 && MW==9) begin:c36x9
    RAMB16_S9_S36 ram (
    .ADDRA(waddr), .DIA(wid),.DIPA(wip), .CLKA(MCI?~wclk:wclk), .ENA(MEI?~wena:wena), .WEA(wiena), .SSRA(L), .DOA(wodb),.DOPA(wopb),
    .ADDRB(vaddr), .DIB(vid),.DIPB(vip), .CLKB(NCI?~vclk:vclk), .ENB(NEI?~vena:vena), .WEB(viena), .SSRB(L), .DOB(vodb),.DOPB(vopb));
  end
  if (NW==36 && MW==18) begin:c36x18
    RAMB16_S18_S36 ram (
    .ADDRA(waddr), .DIA(wid),.DIPA(wip), .CLKA(MCI?~wclk:wclk), .ENA(MEI?~wena:wena), .WEA(wiena), .SSRA(L), .DOA(wodb),.DOPA(wopb),
    .ADDRB(vaddr), .DIB(vid),.DIPB(vip), .CLKB(NCI?~vclk:vclk), .ENB(NEI?~vena:vena), .WEB(viena), .SSRB(L), .DOB(vodb),.DOPB(vopb));
  end
  if (NW==36 && MW==36) begin:c36x36
    RAMB16_S36_S36 ram (
    .ADDRA(vaddr), .DIA(vid),.DIPA(vip), .CLKA(NCI?~vclk:vclk), .ENA(NEI?~vena:vena), .WEA(viena), .SSRA(L), .DOA(vodb),.DOPA(vopb),
    .ADDRB(waddr), .DIB(wid),.DIPB(wip), .CLKB(MCI?~wclk:wclk), .ENB(MEI?~wena:wena), .WEB(wiena), .SSRB(L), .DOB(wodb),.DOPB(wopb));
  end
  endgenerate

`elsif SPARTAN6

  wire [LRS-1:0] vadrx = {vaddr,{LND{L}}};
  wire [LRS-1:0] wadrx = {waddr,{LMD{L}}};
  RAMB16BWER ram (
  .CLKA(NCI?~vclk:vclk),.ENA(NEI?~vena:vena),.WEA(vjena),.RSTA(L), .ADDRA(vadrx), .DIA(vid),.DIPA(vip), .DOA(vod),.DOPA(vop),
  .CLKB(MCI?~wclk:wclk),.ENB(MEI?~wena:wena),.WEB(wjena),.RSTB(L), .ADDRB(wadrx), .DIB(wid),.DIPB(wip), .DOB(wod),.DOPB(wop),
  .REGCEA(NREG),.REGCEB(MREG));
  defparam ram.DOA_REG = NREG;
  defparam ram.DATA_WIDTH_A = NW;
  defparam ram.DOB_REG = MREG; 
  defparam ram.DATA_WIDTH_B = MW;
  defparam ram.SIM_DEVICE = "SPARTAN6";

`elsif NO32BRAM

  wire [LRS:0] vadrx = {H,vaddr,{LND{H}}};
  wire [LRS:0] wadrx = {H,waddr,{LMD{H}}};
  RAMB16 #(.DOA_REG(NREG),.READ_WIDTH_A(NW),.WRITE_WIDTH_A(NW),.WRITE_MODE_A(NWM),
           .DOB_REG(MREG),.READ_WIDTH_B(MW),.WRITE_WIDTH_B(MW),.WRITE_MODE_B(MWM)) ram (
    .CLKA(NCI?~vclk:vclk),.ENA(NEI?~vena:vena),.WEA(vjena),.SSRA(L), .ADDRA(vadrx), .DIA(vid),.DIPA(vip), .DOA(vod),.DOPA(vop),
    .CLKB(MCI?~wclk:wclk),.ENB(MEI?~wena:wena),.WEB(wjena),.SSRB(L), .ADDRB(wadrx), .DIB(wid),.DIPB(wip), .DOB(wod),.DOPB(wop),
    .REGCEA(NREG),.REGCEB(MREG));

`else

  wire [LRS:0] vadrx = {H,vaddr,{LND{H}}};
  wire [LRS:0] wadrx = {H,waddr,{LMD{H}}};
  generate 
  if (LRS<15 && GW<=(SDP?36:18)) begin:r16k
    wire [1:0] xipa,xipb,xopa,xopb;
    wire [15:0] xida,xidb,xoda,xodb;
    assign xida = (SDP && MW==36)? wid[15:0]  : vid;
    assign xipa = (SDP && MW==36)? wip[1:0]   : vip;
    assign xidb = (SDP && MW==36)? wid[31:16] : wid;
    assign xipb = (SDP && MW==36)? wip[3:2]   : wip;
    RAMB18E1 #(.DOA_REG(NREG),.READ_WIDTH_A(NW),.WRITE_WIDTH_A(SDP?0:NW),.WRITE_MODE_A(NWM),
               .DOB_REG(MREG),.READ_WIDTH_B(SDP?0:MW),.WRITE_WIDTH_B(MW),.WRITE_MODE_B(MWM),.RAM_MODE(SDP?"SDP":"TDP")) ram (
    .CLKARDCLK(NCI?~vclk:vclk),.ENARDEN(NEI?~vena:vena),.WEA  (vjena), .ADDRARDADDR(vadrx), .DIADI(xida),.DIPADIP(xipa), .DOADO(xoda),.DOPADOP(xopa),
    .CLKBWRCLK(MCI?~wclk:wclk),.ENBWREN(MEI?~wena:wena),.WEBWE(wjena), .ADDRBWRADDR(wadrx), .DIBDI(xidb),.DIPBDIP(xipb), .DOBDO(xodb),.DOPBDOP(xopb),
    .REGCEAREGCE(NREG),.REGCEB(MREG),.RSTRAMARSTRAM(L),.RSTRAMB(L),.RSTREGARSTREG(L),.RSTREGB(L));
    assign vod = {xodb,xoda}; assign wod = xodb;
    assign vop = {xopb,xopa}; assign wop = xopb;
  end
  else begin:r32k
    wire [3:0] xipa,xipb,xopa,xopb;
    wire [31:0] xida,xidb,xoda,xodb;
    assign xida = (SDP && MW==72)? wid[31:0]  : vid;
    assign xipa = (SDP && MW==72)? wip[3:0]   : vip;
    assign xidb = (SDP && MW==72)? wid[63:32] : wid;
    assign xipb = (SDP && MW==72)? wip[7:4]   : wip;
    RAMB36E1 #(.DOA_REG(NREG),.READ_WIDTH_A(NW),.WRITE_WIDTH_A(SDP?0:NW),.WRITE_MODE_A(NWM),
               .DOB_REG(MREG),.READ_WIDTH_B(SDP?0:MW),.WRITE_WIDTH_B(MW),.WRITE_MODE_B(MWM),.RAM_MODE(SDP?"SDP":"TDP")) ram (
    .CLKARDCLK(NCI?~vclk:vclk),.ENARDEN(NEI?~vena:vena),.WEA  (vjena), .ADDRARDADDR(vadrx), .DIADI(xida),.DIPADIP(xipa), .DOADO(xoda),.DOPADOP(xopa),
    .CLKBWRCLK(MCI?~wclk:wclk),.ENBWREN(MEI?~wena:wena),.WEBWE(wjena), .ADDRBWRADDR(wadrx), .DIBDI(xidb),.DIPBDIP(xipb), .DOBDO(xodb),.DOPBDOP(xopb),
    .REGCEAREGCE(NREG),.REGCEB(MREG),.RSTRAMARSTRAM(L),.RSTRAMB(L),.RSTREGARSTREG(L),.RSTREGB(L));
    assign vod = {xodb,xoda}; assign wod = xodb;
    assign vop = {xopb,xopa}; assign wop = xopb;
  end
  endgenerate

`endif

endmodule




`ifndef verilator

// specialty for jvmcodes lookup tables
module dpram_2048x9x36 (vclk, vce, vaddr, viena, vidat, vodat,
                        wclk, wce, waddr, wiena, widat, wodat);

  input vclk, vce, viena, wclk, wce, wiena;
  input [11:0] vaddr;
  input [8:0] vidat;
  output [8:0] vodat;

  input [9:0] waddr;
  input [35:0] widat;
  output [35:0] wodat;

  wire H=1, L=0;
  RAMB16_S9_S36 inst (
   .CLKA(vclk),.ADDRA(vaddr), .DIA(vidat[7:0]),.DIPA(vidat[8]),      .ENA(vce),.WEA(viena),.SSRA(L), .DOA(vodat[7:0]),.DOPA(vodat[8]),
   .CLKB(wclk),.ADDRB(waddr), .DIB(widat[31:0]),.DIPB(widat[35:32]), .ENB(wce),.WEB(wiena),.SSRB(L), .DOB(wodat[31:0]),.DOPB(wodat[35:32]));
 `include "../lib/jvmlut.hx"

endmodule

// specialty for sin/cos lookup tables
module dpram_4096x16x16 (vclk, vce, vaddr, viena, vidat, vodat,
                         wclk, wce, waddr, wiena, widat, wodat);

  input vclk, vce, viena, wclk, wce, wiena;
  input [11:0] vaddr;
  input [15:0] vidat;
  output [15:0] vodat;
  input [11:0] waddr;
  input [15:0] widat;
  output [15:0] wodat;

  wire H=1, L=0;
  RAMB16_S4_S4 ram0 (.ADDRA(vaddr), .DIA(vidat[3+0:0]), .CLKA(vclk), .ENA(vce), .WEA(viena), .SSRA(L), .DOA(vodat[3+0:0]),
                     .ADDRB(waddr), .DIB(widat[3+0:0]), .CLKB(wclk), .ENB(wce), .WEB(wiena), .SSRB(L), .DOB(wodat[3+0:0]));
  RAMB16_S4_S4 ram1 (.ADDRA(vaddr), .DIA(vidat[3+4:4]), .CLKA(vclk), .ENA(vce), .WEA(viena), .SSRA(L), .DOA(vodat[3+4:4]),
                     .ADDRB(waddr), .DIB(widat[3+4:4]), .CLKB(wclk), .ENB(wce), .WEB(wiena), .SSRB(L), .DOB(wodat[3+4:4]));
  RAMB16_S4_S4 ram2 (.ADDRA(vaddr), .DIA(vidat[3+8:8]), .CLKA(vclk), .ENA(vce), .WEA(viena), .SSRA(L), .DOA(vodat[3+8:8]),
                     .ADDRB(waddr), .DIB(widat[3+8:8]), .CLKB(wclk), .ENB(wce), .WEB(wiena), .SSRB(L), .DOB(wodat[3+8:8]));
  RAMB16_S4_S4 ram3 (.ADDRA(vaddr), .DIA(vidat[3+12:12]), .CLKA(vclk), .ENA(vce), .WEA(viena), .SSRA(L), .DOA(vodat[3+12:12]),
                     .ADDRB(waddr), .DIB(widat[3+12:12]), .CLKB(wclk), .ENB(wce), .WEB(wiena), .SSRB(L), .DOB(wodat[3+12:12]));
 `include "../lib/sinlut.hx"

endmodule

// specialty for FFT twiddle tables
module dpram_1024x32x32 (vclk, vce, vaddr, viena, vidat, vodat,
                         wclk, wce, waddr, wiena, widat, wodat);

  input vclk, vce, viena, wclk, wce, wiena;
  input [9:0] vaddr;
  input [31:0] vidat;
  output [31:0] vodat;
  input [9:0] waddr;
  input [31:0] widat;
  output [31:0] wodat;

  wire H=1, L=0;
  genvar i;
  generate
  for (i=0; i<32; i=i+16) begin:ram
    RAMB16_S18_S18 inst (
    .ADDRA(vaddr), .DIA(vidat[i+15:i]), .CLKA(vclk), .ENA(vce), .WEA(viena), .SSRA(L), .DOA(vodat[i+15:i]),
    .ADDRB(waddr), .DIB(widat[i+15:i]), .CLKB(wclk), .ENB(wce), .WEB(wiena), .SSRB(L), .DOB(wodat[i+15:i]));
  end
  endgenerate

endmodule

// PCI bidirectional channelized fifo (dir=1 vi-to-wo) (dir=0 wi-to-vo)
module bififo128k64x64c4K (vclk,vrst,vchan, vival,vidat, voena,vodat,
                           wclk,wrst,wchan, wival,widat, woena,wodat);
  parameter BM=3;
  input vrst, vclk, vival, voena;
  input wrst, wclk, wival, woena;
  input [1:0] vchan, wchan;

  input [63:0] vidat,widat;
  output [63:0] vodat,wodat;

  reg [8:0] vcnt, wcnt;

  wire vval = vival | voena;
  always @ (posedge vclk or posedge vrst) begin
    if (vrst) vcnt <= 0;
    else if (vval) vcnt <= vcnt+1;
  end

  wire wval = woena | wival;
  always @ (posedge wclk or posedge wrst) begin
    if (wrst) wcnt <= 0;
    else if (wval) wcnt <= wcnt+1;
  end

  wire H=1, L=0;
  genvar i;
  generate
  for (i=0; i<64; i=i+8) begin:ram
   RAMB16_S9_S9 inst (
    .ADDRA({vchan,vcnt}), .CLKA(vclk), .ENA(H), .WEA(vival), .SSRA(L), .DIPA(1'h0), .DIA(vidat[i+7:i]), .DOA(vodat[i+7:i]),
    .ADDRB({wchan,wcnt}), .CLKB(wclk), .ENB(H), .WEB(wival), .SSRB(L), .DIPB(1'h0), .DIB(widat[i+7:i]), .DOB(wodat[i+7:i])
   );
  end
  endgenerate

endmodule


// PCI channelized fifo with prefetch
module bififo128k64x64pfc (vclk,vrst,vchan, virdy,vival,vidat, vordy,voena,vodat,
                           wclk,wrst,wchan, wirdy,wival,widat, wordy,woena,wodat,
                           vopre,usepre,wide,endian);

  input vrst, vclk, vival, voena, vopre, usepre, wide;
  input wrst, wclk, wival, woena;
  input [3:0] vchan, wchan;
  input [2:0] endian;
  output virdy, vordy, wirdy, wordy;

  input [63:0] vidat,widat;
  output [63:0] vodat,wodat;

  reg prime,primed,voenay,vopred,prep,vohigh,vosel,visel,vivalx;
  wire voenaz = (prime || voena);
  wire voenax = (prime || voena) && (wide || vohigh);
  wire [63:0] vodatx;
  reg [63:0] vidatx;
  reg [1:0] pcnt;

  // generate 3 prime cycles
  always @ (posedge vclk or posedge vrst) begin
    if (vrst) primed <= 0;
    else primed <= (pcnt==1) || primed;
    // need 4 prime cycles triggered by reset
    if (vrst) prime <= 0;
    else prime <= (prep||prime) && !primed;
    // count of data in prefetch buffer
    if (vrst) pcnt <= 0;
    else if (voenay && !voena) pcnt <= pcnt+1;
    else if (!voenay && voena) pcnt <= pcnt-1;
    if (vrst) vohigh <= 0;
    else if (voenaz) vohigh <= vohigh ^ (!wide);
    if (vrst) visel <= 0;
    else if (vival) visel <= visel ^ (!wide);
  end
  always @ (posedge vclk) begin
    voenay <= voenaz;
    vopred <= vopre;
    vosel  <= vohigh;
    prep   <= vrst && usepre;
    vivalx <= vival && (wide || visel);
    if (vival) vidatx <= wide? vidat : {vidat[31:0],vidatx[63:32]};
  end

//  wire [1:0] padj = vopre? 1 : vopred? 2 : 3;
  wire [1:0] padj = {!vopre,!vopred};
  wire [1:0] addr = pcnt-padj;

  assign virdy = 1;
  assign vordy = 1;
  assign wirdy = 1;
  reg wordy;
  always @(posedge wclk) begin
    wordy <= (vchan!=wchan);
  end

  bififo128k64x64c4K ff (vclk,vrst,vchan[1:0], vivalx,vidatx, voenax,vodatx,
                         wclk,wrst,wchan[1:0], wival,widat, woena,wodat);

  srl16xN #(64) sr (vclk, voenay, {2'b00,addr}, {vodatx[63:32],vosel?vodatx[63:32]:vodatx[31:0]}, vodat);

endmodule

`endif

module cram64xNs (q, d, a, clk,we);
  parameter N=1;
  input clk,we;
  input [N-1:0] d;
  input [5:0] a;
  output [N-1:0] q;

`ifdef VIVADO
  genvar i;
  generate
  for (i=0; i<N; i=i+1) begin:rb
    RAM64X1S cr (.WCLK(clk),.WE(we),.D(d[i]),.A0(a[0]),.A1(a[1]),.A2(a[2]),.A3(a[3]),.A4(a[4]),.A5(a[5]),.O(q[i]));
  end
  endgenerate
`else
  //Synthesis attribute ram_style of ram is distributed
  reg [N-1:0] ram [63:0] /* synthesis syn_ramstyle="select_ram" */;
  always @(posedge clk) if (we) ram[a] <= d;
  assign q = ram[a];   
`endif

endmodule

module cram16xNs (q, d, a, clk,we);
  parameter N=1;
  input clk,we;
  input [N-1:0] d;
  input [3:0] a;
  output [N-1:0] q;

`ifdef VIVADO
  genvar i;
  generate
  for (i=0; i<N; i=i+1) begin:rb
    RAM32X1S cr (.WCLK(clk),.WE(we),.D(d[i]),.A0(a[0]),.A1(a[1]),.A2(a[2]),.A3(a[3]),.A4(1'b0),.O(q[i]));
  end
  endgenerate
`else
  //Synthesis attribute ram_style of ram is distributed
  reg [N-1:0] ram [15:0] /* synthesis syn_ramstyle="select_ram" */;
  always @(posedge clk) if (we) ram[a] <= d;
  assign q = ram[a];   
`endif

endmodule

module cram16xNd (q, d, a, clk,we, ar,qr);
  parameter N=1;
  input clk,we;
  input [N-1:0]  d;
  input [3:0] a,ar;
  output [N-1:0] q,qr;

`ifdef VIVADO
  genvar i;
  generate
  for (i=0; i<N; i=i+1) begin:rb
    RAM32X1D cr (.WCLK(clk),.WE(we),.D(d[i]),.A0(a[0]),.A1(a[1]),.A2(a[2]),.A3(a[3]),.A4(1'b0),.SPO(q[i]),
            .DPRA0(ar[0]),.DPRA1(ar[1]),.DPRA2(ar[2]),.DPRA3(ar[3]),.DPRA4(1'b0),.DPO(qr[i]));
  end
  endgenerate
`else
  //Synthesis attribute ram_style of ram is distributed
  reg [N-1:0] ram [15:0] /* synthesis syn_ramstyle="select_ram" */;
  always @(posedge clk) if (we) ram[a] <= d;
  assign q = ram[a];   
  assign qr = ram[ar];   
`endif

endmodule

module cram16xNsd (d,a, clk,we, ar,qr);
  parameter N=1;
  input clk,we;
  input  [N-1:0] d;
  input [3:0] a,ar;
  output [N-1:0] qr;

  wire [N-1:0] q;
  cram16xNd #(N) cc (q,d,a, clk,we, ar,qr);

endmodule

module cram16xNsdp (d,a, clk,we, ar,qr);
  parameter N=1;
  input clk,we;
  input  [N-1:0] d;
  input [3:0] a,ar;
  output reg [N-1:0] qr;

  wire [N-1:0] q,pr;
  cram16xNd #(N) cc (q,d,a, clk,we, ar,pr);
  always @(posedge clk) qr <= pr;

endmodule

/* dual port with asynchronous read - (for sparse we only) */
module cram16xNda (q, d, a, clk,we, ar,qr,clkr);
  parameter N=1;
  input clk,we,clkr;
  input [N-1:0] d;
  input [3:0] a,ar;
  output [N-1:0] q,qr;

  wire [N-1:0] pr;
  cram16xNd #(N) cc (q,d,a, clk,we, ar,pr);

  reg [3:0] aw;
  reg [N-1:0] dw;
  reg cw,cr,re;

  always @(posedge clk) if (we) aw <= a;
  always @(posedge clk) if (we) dw <= d;
  always @(posedge clk) if (we) cw <= !cw;

  always @(posedge clkr) re <= (cw^cr) && !re;
  always @(posedge clkr) if (re) cr <= !cr;

  wire [N-1:0] qw;
  cram16xNd #(N) dd (qw,dw,aw, clkr,re, ar,qr);

endmodule


module xfifo36 (rst, vclk,vrdy,vena,vdat,
                     wclk,wrdy,wena,wdat);
  parameter AFULL=480;
  parameter AEMPTY=16;

  localparam AFO=504;  // workaround for async flag problems - bad if this ever crosses
  localparam AEO=500;  // workaround for async flag problems - bad if this ever crosses

  input rst,vclk,vena,wclk,wena;
  input [35:0] vdat;
  output [35:0] wdat;
  output vrdy,wrdy;

  wire empty;
  wire [8:0] wcnt,rcnt;

  FIFO16 #(.DATA_WIDTH(36), .FIRST_WORD_FALL_THROUGH("TRUE"), .ALMOST_FULL_OFFSET(AFO), .ALMOST_EMPTY_OFFSET(AEO) ) 
  fifo ( 
    .RST(rst),    .EMPTY(empty), .RDCOUNT(rcnt), .WRCOUNT(wcnt), .FULL(),
    .WRCLK(vclk), .WREN(vena), .DIP(vdat[35:32]), .DI(vdat[31:0]), 
    .RDCLK(wclk), .RDEN(wena), .DOP(wdat[35:32]), .DO(wdat[31:0]) 
  );

  reg afull1,afull2,afull;
  reg [8:0] diff;
  always @(posedge vclk) begin
    diff  <= (wcnt-rcnt);
    afull1 <= (diff>AFULL);
    afull2 <= afull1;
    afull  <= afull1 && afull2;
  end

  assign vrdy = !afull;
  assign wrdy = !empty;

endmodule

// data formats 0=SI 1=SB 2=SN 3=SP 4=SL 5=SX   8=CI 9=CB C=CL
        
// packed data FIFO with output reformat to wofmt=CI 
module fifoNxMfmt (vclk,vrst, vifmt,virdy,vival,vidat, 
                   wclk,wrst, wofmt,wordy,woena,wodat);
  parameter N=64;	// input width
  parameter M=32;	// output width per channel (always 32)
  parameter DP=0;	// num DP rams
  parameter NC=1;	// number output channels
  input vrst, vclk, vival;
  input wrst, wclk, woena;
  output virdy, wordy;
  input [N-1:0] vidat;
  output [NC*M-1:0] wodat;
  input [3:0] vifmt;
  input [4:0] wofmt;

  wire H=1, L=0;
  wire vordy,voena,wirdy,wival,xordy,xoena;
  wire [N-1:0] vodat;
  wire [NC*M-1:0] widat,xodat;

  bififoNxM #(N,NC*M,DP,3,`WONLY,`RONLY|`CENA) ff 
   (vclk,vrst, virdy,vival,vidat, vordy,voena,vodat, 
    wclk,wrst, wirdy,wival,widat, xordy,xoena,xodat, H);

  reg [2:0] cnt;
  reg [3:0] clr;
  reg prime,next1,next2;
  assign wordy = xordy && !prime;			// output ready
  wire ena = (prime && xordy) || woena;			// count/selector enable
  wire dbl = wofmt[4];	// double wide with NC=2
  wire next;
  assign xoena = (prime && xordy) || (woena && next);	// fifo read signal

  wire sb = (vifmt==1);
  wire si = (vifmt==0);
  wire cb = (vifmt==9);
  wire ci = (vifmt==8);
  
  always @(posedge wclk) begin
    // count is one cycle early (uses prime) to latch select in output cycle
    if (wrst) cnt <= 0; else if (ena) cnt <= cnt+1;
    // prime will be up for one cycle with xordy to pipeline the fifo output
    if (wrst) prime <= 1; else prime <= prime && !xordy;
    // which bytes to clear
    clr <= (vifmt==1)? 4'b1101 : (vifmt==0)? 4'b1100 : (vifmt==9)? 4'b0101 : 4'b0000;
    // next is when we need to read the next fifo entry 1=SB 0=SI 9=CB x=CI
    if (ena) next1 <= sb? (cnt[1:0]==3) : si? cnt[0] :      cb? cnt[0]      : H;
    if (ena) next2 <= sb? (cnt[2:0]==7) : si? cnt[1:0]==3 : cb? cnt[1:0]==3 : cnt[0];
  end

  wire c0 = cnt[0];
  wire c1 = cnt[1];
  wire c2 = cnt[2];
  reg [NC:0] sel [4*NC-1:0];	// the selects pick the fifo byte to present at the output

  generate
  if (NC==2) begin:NC2
    assign next = dbl? next1 : next2;
    always @(posedge wclk) begin
      if (ena) sel[7] <= dbl? (                               cb? {c0,H,H} : 7) : (                          cb? {c1,c0,H} : {c0,H,H});
      if (ena) sel[6] <= dbl? (                                              6) : (                                          {c0,H,L});
      if (ena) sel[5] <= dbl? (sb? {c1,c0,H} : si? {c0,H,H} : cb? {c0,H,L} : 5) : (sb? cnt : si? {c1,c0,H} : cb? {c1,c0,L} : {c0,L,H});
      if (ena) sel[4] <= dbl? (                si? {c0,H,L} :                4) : (          si? {c1,c0,L} :                 {c0,L,L});
      if (ena) sel[3] <= dbl? (                               cb? {c0,L,H} : 3) : (                          cb? {c1,c0,H} : {c0,H,H});
      if (ena) sel[2] <= dbl? (                                              2) : (                                          {c0,H,L});
      if (ena) sel[1] <= dbl? (sb? {c1,c0,L} : si? {c0,L,H} : cb? {c0,L,L} : 1) : (sb? cnt : si? {c1,c0,H} : cb? {c1,c0,L} : {c0,L,H});
      if (ena) sel[0] <= dbl? (                si? {c0,L,L} :                0) : (          si? {c1,c0,L} :                 {c0,L,L});
    end
    muxMxNpec #(8,8) m7 (wodat[63:56], xodat, sel[7],wclk,woena,clr[3]);
    muxMxNpec #(8,8) m6 (wodat[55:48], xodat, sel[6],wclk,woena,clr[2]);
    muxMxNpec #(8,8) m5 (wodat[47:40], xodat, sel[5],wclk,woena,clr[1]);
    muxMxNpec #(8,8) m4 (wodat[39:32], xodat, sel[4],wclk,woena,clr[0]);
    muxMxNpec #(8,8) m3 (wodat[31:24], xodat, sel[3],wclk,woena,clr[3]);
    muxMxNpec #(8,8) m2 (wodat[23:16], xodat, sel[2],wclk,woena,clr[2]);
    muxMxNpec #(8,8) m1 (wodat[15:08], xodat, sel[1],wclk,woena,clr[1]);
    muxMxNpec #(8,8) m0 (wodat[07:00], xodat, sel[0],wclk,woena,clr[0]);
  end
  if (NC==1) begin:NC1
    assign next = next1;
    always @(posedge wclk) begin	
      if (ena) sel[3] <=                        cb? {c0,H} : 3;
      if (ena) sel[2] <=                                     2;
      if (ena) sel[1] <= sb? cnt : si? {c0,H} : cb? {c0,L} : 1;
      if (ena) sel[0] <=           si? {c0,L}              : 0;
    end
    muxMxNpec #(4,8) m3 (wodat[31:24], xodat, sel[3],wclk,woena,clr[3]);
    muxMxNpec #(4,8) m2 (wodat[23:16], xodat, sel[2],wclk,woena,clr[2]);
    muxMxNpec #(4,8) m1 (wodat[15:08], xodat, sel[1],wclk,woena,clr[1]);
    muxMxNpec #(4,8) m0 (wodat[07:00], xodat, sel[0],wclk,woena,clr[0]);
  end
  endgenerate

endmodule

// packed data FIFO with output reformat from wifmt=CI|CL
module fifoNxMfmti (vclk,vrst, vofmt,vordy,voena,vodat, 
                    wclk,wrst, wifmt,wirdy,wival,widat);
  parameter N=64;
  parameter M=32;
  parameter DP=2;
`ifdef verilator
  parameter BM=6;
`else
  parameter BM=3;
`endif
  input vrst, vclk, voena;
  input wrst, wclk, wival;
  output vordy, wirdy;
  output [N-1:0] vodat;
  input [M-1:0] widat;
  input [3:0] vofmt, wifmt;

  wire H=1, L=0;
  wire vival,woena,virdy,wordy;
  wire [N-1:0] vidat;
  wire [M-1:0] wodat,xidat;
  reg xival;

`ifdef MCS_FMTI
  bififoNxM #(N,M,DP,BM,`RONLY|`DBLBUF,`WONLY) ff
`else
  bififoNxM #(N,M,DP,BM,`RONLY,`WONLY) ff
`endif
   (vclk,vrst, virdy,vival,vidat, vordy,voena,vodat, 
    wclk,wrst, wirdy,xival,xidat, wordy,woena,wodat, L);

  reg [1:0] cnt;
  reg [7:0] sel;

 generate

 if (M==32) begin:CI
  always @(posedge wclk) begin // here to stop verilator overoptimization
    xival <= wival && ((vofmt==1)? (cnt==3) : (vofmt==0)? cnt[0] : (vofmt==9)? cnt[0] : H);
    if (wrst) cnt <= 0; else if (wival) cnt <= cnt+1;
  end
  always @(posedge wclk) begin
    sel[7:6] <= (vofmt==1)? 1 : (vofmt==0)? 1 : (vofmt==9)? 3 : (vofmt==4)? 1 : 3;
    sel[5:4] <= (vofmt==1)? 1 : (vofmt==0)? 0 : (vofmt==9)? 1 : (vofmt==4)? 0 : 2;
    sel[3:2] <= (vofmt==1)? 1 : (vofmt==0)? 1 : (vofmt==9)? 3 : (vofmt==4)? 3 : 1;
    sel[1:0] <= (vofmt==1)? 1 : (vofmt==0)? 0 : (vofmt==9)? 1 : (vofmt==4)? 2 : 0;
  end
  wire ena3 = wival && ((vofmt==1)? (cnt==3) : (vofmt==0)?  cnt[0] : (vofmt==9)?  cnt[0] : H);
  wire ena2 = wival && ((vofmt==1)? (cnt==2) : (vofmt==0)?  cnt[0] : (vofmt==9)?  cnt[0] : H);
  wire ena1 = wival && ((vofmt==1)? (cnt==1) : (vofmt==0)? !cnt[0] : (vofmt==9)? !cnt[0] : H);
  wire ena0 = wival && ((vofmt==1)? (cnt==0) : (vofmt==0)? !cnt[0] : (vofmt==9)? !cnt[0] : H);

  mux4xNpe #(8) m3 (xidat[31:24], widat[7:0],widat[15:8],widat[23:16],widat[31:24], sel[7:6],wclk,ena3);
  mux4xNpe #(8) m2 (xidat[23:16], widat[7:0],widat[15:8],widat[23:16],widat[31:24], sel[5:4],wclk,ena2);
  mux4xNpe #(8) m1 (xidat[15:08], widat[7:0],widat[15:8],widat[23:16],widat[31:24], sel[3:2],wclk,ena1);
  mux4xNpe #(8) m0 (xidat[07:00], widat[7:0],widat[15:8],widat[23:16],widat[31:24], sel[1:0],wclk,ena0);
 end

 if (M==64) begin:CL
  always @(posedge wclk) begin // here to stop verilator overoptimization
    xival <= wival && ((vofmt==0)? (cnt==3) : (vofmt==4)? cnt[0] : (vofmt==8)? cnt[0] : H);
    if (wrst) cnt <= 0; else if (wival) cnt <= cnt+1;
  end
  always @(posedge wclk) begin
    sel[7:6] <= (vofmt==0)? 1 : (vofmt==4)? 1 : (vofmt==8)? 3 : (vofmt==4)? 1 : 3;
    sel[5:4] <= (vofmt==0)? 1 : (vofmt==4)? 0 : (vofmt==8)? 1 : (vofmt==4)? 0 : 2;
    sel[3:2] <= (vofmt==0)? 1 : (vofmt==4)? 1 : (vofmt==8)? 3 : (vofmt==4)? 3 : 1;
    sel[1:0] <= (vofmt==0)? 1 : (vofmt==4)? 0 : (vofmt==8)? 1 : (vofmt==4)? 2 : 0;
  end
  wire ena3 = wival && ((vofmt==0)? (cnt==3) : (vofmt==4)?  cnt[0] : (vofmt==8)?  cnt[0] : H);
  wire ena2 = wival && ((vofmt==0)? (cnt==2) : (vofmt==4)?  cnt[0] : (vofmt==8)?  cnt[0] : H);
  wire ena1 = wival && ((vofmt==0)? (cnt==1) : (vofmt==4)? !cnt[0] : (vofmt==8)? !cnt[0] : H);
  wire ena0 = wival && ((vofmt==0)? (cnt==0) : (vofmt==4)? !cnt[0] : (vofmt==8)? !cnt[0] : H);

  mux4xNpe #(16) m3 (xidat[63:48], widat[15:0],widat[31:16],widat[47:32],widat[63:48], sel[7:6],wclk,ena3);
  mux4xNpe #(16) m2 (xidat[47:32], widat[15:0],widat[31:16],widat[47:32],widat[63:48], sel[5:4],wclk,ena2);
  mux4xNpe #(16) m1 (xidat[31:16], widat[15:0],widat[31:16],widat[47:32],widat[63:48], sel[3:2],wclk,ena1);
  mux4xNpe #(16) m0 (xidat[15:00], widat[15:0],widat[31:16],widat[47:32],widat[63:48], sel[1:0],wclk,ena0);
 end

  endgenerate


endmodule

module swrstat (clk,cs,wr, addr,dati,dato);
  parameter PARAM0=0, PARAM1=0, PARAM2=0, PARAM3=0, PARAM4=0, PARAM5=0, PARAM6=0, PARAM7=`SYS_CLKF;
  input clk,cs,wr;
  input [31:0] addr,dati;
  output [31:0] dato;
  wire   [31:0] dummy;
  wire L=0,H=1;
  swrstatdp #(PARAM0, PARAM1, PARAM2, PARAM3, PARAM4, PARAM5, PARAM6, PARAM7) 
        swdp (clk,cs,wr, addr,dati,dato, clk,L,L, 32'd0,32'd0,dummy);

endmodule

module swrstatdbg (clk,cs,wr, addr,dati,dato, dbgclk,dbgbus);
  parameter PARAM0=0, PARAM1=0, PARAM2=0, PARAM3=0, PARAM4=0, PARAM5=0, PARAM6=0, PARAM7=`SYS_CLKF, DBGBW=32;
  input clk,cs,wr;
  input [31:0] addr,dati;
  output [31:0] dato;
  input dbgclk;
  input [31:0] dbgbus;
  wire  [31:0] dummy,datos,datod;
  wire L=0,H=1;

  swrstatdp #(PARAM0, PARAM1, PARAM2, PARAM3, PARAM4, PARAM5, PARAM6, PARAM7) 
        swdp (clk,cs,wr, addr,dati,datos, clk,L,L, 32'd0,32'd0,dummy);

  reg dbgsys;
  reg [7:0] dbgctl;
  wire reset = !dbgctl[0];	// dbg reset
  wire oneshot = dbgctl[1];	// dbg oneshot
  wire gatem = dbgctl[2];	// dbg gate
  wire edged = dbgctl[3];	// dbg edge
  always @(posedge clk) begin
    dbgsys  <= cs && wr && addr[9:2]==0;
    if (dbgsys) dbgctl <= dati[31:24];
  end
  reg [3:0] dbgsel;
  reg rst,active,gated;
  reg [10:0] dbgcnt;
  reg [DBGBW-1:0] dbgbusd,dbgbuse;
  wire dbgbusi = dbgbusd[dbgsel];
  wire done = (oneshot && dbgcnt[10]);
  wire dbgwr = active && !(oneshot && done) && !gated;
  always @(posedge dbgclk) begin
    rst <= reset;
    dbgsel <= dbgctl[7:4];
    dbgbusd <= dbgbus;
    dbgbuse <= dbgbusd;
    if (rst) active <= 0; else active <= dbgbusi || (active && !done);
    if (rst) dbgcnt <= 0; else if (dbgwr) dbgcnt <= dbgcnt+1;
    gated <= gatem && !dbgbusi;
  end
  sdpram #(4,DBGBW,32) dbgr (dbgclk,H, dbgcnt,dbgwr, dbgbuse,   clk,cs, addr[11:2],datod);

  assign dato = addr[12]? datod : datos;	// debug is upper 4K

endmodule


module swrstatdp (clk1,cs1,wr1, addr1,dati1,dato1, clk2,cs2,wr2, addr2,dati2,dato2);
  parameter PARAM0=0, PARAM1=0, PARAM2=0, PARAM3=0, PARAM4=0, PARAM5=0, PARAM6=0, PARAM7=0;
  localparam P0B0=(PARAM0>>0)&255;
  localparam P0B1=(PARAM0>>8)&255;
  localparam P0B2=(PARAM0>>16)&255;
  localparam P0B3=(PARAM0>>24)&255;	// need to flip bytes around for strange char handling
  localparam P0=(P0B2==0)? (P0B0<<8)|(P0B1<<0) : (P0B3==0)? (P0B0<<16)|(P0B1<<8)|(P0B2<<0) : (P0B0<<24)|(P0B1<<16)|(P0B2<<8)|(P0B3<<0);

  input clk1,cs1,wr1,clk2,cs2,wr2;
  input [31:0] addr1,dati1,addr2,dati2;
  output [31:0] dato1,dato2;

  function [255:0] collect; // needed to make a constant of the input parameters
    input [31:0] param0,param1,param2,param3,param4,param5,param6,param7;
    // param0 is byte swapped since it comes in as a verilog string this way
    collect = {param7,param6,param5,param4,param3,param2,param1,param0};
  endfunction

  RAMB16_S36_S36 ram ( .ADDRA(addr1[10:2]), .DIA(dati1),.DIPA(4'b0), .CLKA(clk1), .ENA(cs1), .WEA(wr1), .SSRA(1'b0), .DOA(dato1), .DOPA(),
                       .ADDRB(addr2[10:2]), .DIB(dati2),.DIPB(4'b0), .CLKB(clk2), .ENB(cs2), .WEB(wr2), .SSRB(1'b0), .DOB(dato2), .DOPB());
  defparam ram.INIT_3F = collect(P0,PARAM1,PARAM2,PARAM3,PARAM4,PARAM5,PARAM6,PARAM7);

endmodule

